@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2014-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,9 +22,11 @@
 #include <sys/mman.h>
 #include <sys/time.h>
 
+#if !defined(KODEV_MINIMAL)
 #include <openssl/err.h>
 #include <openssl/pem.h>
 #include <openssl/x509v3.h>
+#endif
 
 #include <ctype.h>
 #include <errno.h>
@@ -140,7 +142,7 @@ static void		fatal(const char *, ...) __attribute__((noreturn));
 
 static void		cli_file_close(int);
 static void		cli_run_kore(void);
-static void		cli_generate_certs(void);
+static void		cli_run_kore_python(void);
 static void		cli_compile_kore(void *);
 static void		cli_link_application(void *);
 static void		cli_compile_source_file(void *);
@@ -163,7 +165,6 @@ static void		cli_write_asset(const char *, const char *,
 			    struct buildopt *);
 static void		cli_register_kore_file(char *, struct dirent *);
 static void		cli_register_source_file(char *, struct dirent *);
-static void		cli_file_create(const char *, const char *, size_t);
 static int		cli_file_requires_build(struct stat *, const char *);
 static void		cli_find_files(const char *,
 			    void (*cb)(char *, struct dirent *));
@@ -196,15 +197,22 @@ static void		cli_help(int, char **);
 static void		cli_info(int, char **);
 static void		cli_build(int, char **);
 static void		cli_clean(int, char **);
-static void		cli_create(int, char **);
 static void		cli_reload(int, char **);
 static void		cli_flavor(int, char **);
 
+#if !defined(KODEV_MINIMAL)
+static void		cli_create(int, char **);
 static void		cli_create_help(void);
 
 static void		file_create_src(void);
 static void		file_create_config(void);
 static void		file_create_gitignore(void);
+static void		file_create_python_src(void);
+static void		file_create_python_config(void);
+
+static void		cli_generate_certs(void);
+static void		cli_file_create(const char *, const char *, size_t);
+#endif
 
 static struct cmd cmds[] = {
 	{ "help",	"this help text",			cli_help },
@@ -213,11 +221,14 @@ static struct cmd cmds[] = {
 	{ "info",	"show info on kore on this system",	cli_info },
 	{ "build",	"build an application",			cli_build },
 	{ "clean",	"cleanup the build files",		cli_clean },
+#if !defined(KODEV_MINIMAL)
 	{ "create",	"create a new application skeleton",	cli_create },
+#endif
 	{ "flavor",	"switch between build flavors",		cli_flavor },
 	{ NULL,		NULL,					NULL }
 };
 
+#if !defined(KODEV_MINIMAL)
 static struct filegen gen_files[] = {
 	{ file_create_src },
 	{ file_create_config },
@@ -233,6 +244,18 @@ static const char *gen_dirs[] = {
 	NULL
 };
 
+static const char *python_gen_dirs[] = {
+	"cert",
+	NULL
+};
+
+static struct filegen python_gen_files[] = {
+	{ file_create_python_src },
+	{ file_create_python_config },
+	{ file_create_gitignore },
+	{ NULL }
+};
+
 static const char *http_serveable_function =
 	"int\n"
 	"asset_serve_%s_%s(struct http_request *req)\n"
@@ -258,16 +281,21 @@ static const char *src_data =
 static const char *config_data =
 	"# %s configuration\n"
 	"\n"
-	"bind\t\t127.0.0.1 8888\n"
+	"server tls {\n"
+	"\tbind 127.0.0.1 8888\n"
+	"}\n"
+	"\n"
 	"load\t\t./%s.so\n"
 	"\n"
 	"tls_dhparam\tdh2048.pem\n"
 	"\n"
 	"domain * {\n"
+	"\tattach\t\ttls\n"
+	"\n"
 	"\tcertfile\tcert/server.pem\n"
 	"\tcertkey\t\tcert/key.pem\n"
 	"\n"
-	"\tstatic\t/\tpage\n"
+	"\troute\t/\tpage\n"
 	"}\n";
 
 static const char *build_data =
@@ -306,6 +334,42 @@ static const char *build_data =
 	"#	included if you build with the \"prod\" flavor.\n"
 	"#}\n";
 
+static const char *python_config_data =
+	"# %s configuration\n"
+	"\n"
+	"server tls {\n"
+	"\tbind 127.0.0.1 8888\n"
+	"}\n"
+	"tls_dhparam\tdh2048.pem\n"
+	"\n"
+	"domain * {\n"
+	"\tattach\t\ttls\n"
+	"\n"
+	"\tcertfile\tcert/server.pem\n"
+	"\tcertkey\t\tcert/key.pem\n"
+	"\n"
+	"\troute\t/\tkoreapp.index\n"
+	"}\n";
+
+static const char *python_init_data =
+	"from .app import koreapp\n";
+
+static const char *python_app_data =
+	"import kore\n"
+	"\n"
+	"class App:\n"
+	"    def __init__(self):\n"
+	"        pass\n"
+	"\n"
+	"    def configure(self, args):\n"
+	"        kore.config.file = \"kore.conf\"\n"
+	"        kore.config.deployment = \"development\"\n"
+	"\n"
+	"    async def index(self, req):\n"
+	"        req.response(200, b'')\n"
+	"\n"
+	"koreapp = App()";
+
 static const char *dh2048_data =
 	"-----BEGIN DH PARAMETERS-----\n"
 	"MIIBCAKCAQEAn4f4Qn5SudFjEYPWTbUaOTLUH85YWmmPFW1+b5bRa9ygr+1wfamv\n"
@@ -318,6 +382,8 @@ static const char *dh2048_data =
 
 static const char *gitignore = "*.o\n.flavor\n.objs\n%s.so\nassets.h\ncert\n";
 
+#endif /* !KODEV_MINIMAL */
+
 static int			s_fd = -1;
 static char			*appl = NULL;
 static int			run_after = 0;
@@ -347,6 +413,9 @@ usage(void)
 	int		i;
 
 	fprintf(stderr, "Usage: kodev [command]\n");
+#if defined(KODEV_MINIMAL)
+	fprintf(stderr, "minimal (only build commands supported)\n");
+#endif
 	fprintf(stderr, "\nAvailable commands:\n");
 
 	for (i = 0; cmds[i].name != NULL; i++)
@@ -405,6 +474,7 @@ cli_help(int argc, char **argv)
 	usage();
 }
 
+#if !defined(KODEV_MINIMAL)
 static void
 cli_create_help(void)
 {
@@ -412,6 +482,8 @@ cli_create_help(void)
 	printf("Synopsis:\n");
 	printf("  Create a new application skeleton directory structure.\n");
 	printf("\n");
+	printf("  Optional flags:\n");
+	printf("\t-p = generate a python application skeleton\n");
 
 	exit(1);
 }
@@ -419,16 +491,21 @@ cli_create_help(void)
 static void
 cli_create(int argc, char **argv)
 {
-	int			i, ch;
 	char			*fpath;
 	const char		**dirs;
 	struct filegen		*files;
+	int			i, ch, python;
+
+	python = 0;
 
 	while ((ch = getopt(argc, argv, "hp")) != -1) {
 		switch (ch) {
 		case 'h':
 			cli_create_help();
 			break;
+		case 'p':
+			python = 1;
+			break;
 		default:
 			cli_create_help();
 			break;
@@ -444,8 +521,13 @@ cli_create(int argc, char **argv)
 	appl = argv[0];
 	cli_mkdir(appl, 0755);
 
-	dirs = gen_dirs;
-	files = gen_files;
+	if (python) {
+		dirs = python_gen_dirs;
+		files = python_gen_files;
+	} else {
+		dirs = gen_dirs;
+		files = gen_files;
+	}
 
 	for (i = 0; dirs[i] != NULL; i++) {
 		(void)cli_vasprintf(&fpath, "%s/%s", appl, dirs[i]);
@@ -465,6 +547,7 @@ cli_create(int argc, char **argv)
 	printf("WARNING: DO NOT USE THE GENERATED DH PARAMETERS "
 	    "AND CERTIFICATES IN PRODUCTION\n");
 }
+#endif
 
 static void
 cli_flavor(int argc, char **argv)
@@ -508,13 +591,17 @@ cli_flavor(int argc, char **argv)
 static void
 cli_build(int argc, char **argv)
 {
+#if !defined(KODEV_MINIMAL)
+	int			l;
+	char			*data;
+#endif
 	struct dirent		dp;
 	struct cfile		*cf;
 	struct buildopt		*bopt;
 	struct timeval		times[2];
 	char			*build_path;
-	int			requires_relink, l;
-	char			*sofile, *config, *data;
+	int			requires_relink;
+	char			*sofile, *config;
 	char			*assets_path, *p, *src_path;
 	char			pwd[PATH_MAX], *assets_header;
 
@@ -550,11 +637,14 @@ cli_build(int argc, char **argv)
 
 	cli_flavor_load();
 	bopt = cli_buildopt_new("_default");
+
+#if !defined(KODEV_MINIMAL)
 	if (!cli_file_exists(build_path)) {
 		l = cli_vasprintf(&data, build_data, appl);
 		cli_file_create("conf/build.conf", data, l);
 		free(data);
 	}
+#endif
 
 	cli_find_files(src_path, cli_register_source_file);
 	free(src_path);
@@ -631,6 +721,7 @@ cli_build(int argc, char **argv)
 
 	free(assets_header);
 
+#if !defined(KODEV_MINIMAL)
 	if (bopt->kore_flavor == NULL ||
 	    !strstr(bopt->kore_flavor, "NOTLS=1")) {
 		if (!cli_dir_exists("cert")) {
@@ -638,6 +729,7 @@ cli_build(int argc, char **argv)
 			cli_generate_certs();
 		}
 	}
+#endif
 
 	if (bopt->single_binary) {
 		requires_relink++;
@@ -684,6 +776,11 @@ cli_clean(int argc, char **argv)
 static void
 cli_run(int argc, char **argv)
 {
+	if (cli_file_exists("__init__.py")) {
+		cli_run_kore_python();
+		return;
+	}
+
 	run_after = 1;
 	cli_build(argc, argv);
 
@@ -748,6 +845,34 @@ cli_info(int argc, char **argv)
 	}
 }
 
+#if !defined(KODEV_MINIMAL)
+static void
+file_create_python_src(void)
+{
+	char		*name;
+
+	(void)cli_vasprintf(&name, "%s/__init__.py", appl);
+	cli_file_create(name, python_init_data, strlen(python_init_data));
+	free(name);
+
+	(void)cli_vasprintf(&name, "%s/app.py", appl);
+	cli_file_create(name, python_app_data, strlen(python_app_data));
+	free(name);
+}
+
+static void
+file_create_python_config(void)
+{
+	int		l;
+	char		*name, *data;
+
+	(void)cli_vasprintf(&name, "%s/kore.conf", appl);
+	l = cli_vasprintf(&data, python_config_data, appl);
+	cli_file_create(name, data, l);
+	free(name);
+	free(data);
+}
+
 static void
 file_create_src(void)
 {
@@ -789,6 +914,7 @@ file_create_gitignore(void)
 	free(name);
 	free(data);
 }
+#endif
 
 static void
 cli_mkdir(const char *fpath, int mode)
@@ -928,6 +1054,7 @@ cli_file_write(int fd, const void *buf, size_t len)
 	}
 }
 
+#if !defined(KODEV_MINIMAL)
 static void
 cli_file_create(const char *name, const char *data, size_t len)
 {
@@ -939,6 +1066,7 @@ cli_file_create(const char *name, const char *data, size_t len)
 
 	printf("created %s\n", name);
 }
+#endif
 
 static void
 cli_write_asset(const char *n, const char *e, struct buildopt *bopt)
@@ -946,7 +1074,10 @@ cli_write_asset(const char *n, const char *e, struct buildopt *bopt)
 	cli_file_writef(s_fd, "extern const u_int8_t asset_%s_%s[];\n", n, e);
 	cli_file_writef(s_fd, "extern const u_int32_t asset_len_%s_%s;\n", n, e);
 	cli_file_writef(s_fd, "extern const time_t asset_mtime_%s_%s;\n", n, e);
+
+#if !defined(KODEV_MINIMAL)
 	cli_file_writef(s_fd, "extern const char *asset_sha256_%s_%s;\n", n, e);
+#endif
 
 	if (bopt->flavor_nohttp == 0) {
 		cli_file_writef(s_fd,
@@ -957,17 +1088,21 @@ cli_write_asset(const char *n, const char *e, struct buildopt *bopt)
 static void
 cli_build_asset(char *fpath, struct dirent *dp)
 {
+	u_int8_t		*d;
 	struct stat		st;
+#if !defined(KODEV_MINIMAL)
 	SHA256_CTX		sctx;
+	int			i, len;
+	struct mime_type	*mime;
+	const char		*mime_type;
+	u_int8_t		digest[SHA256_DIGEST_LENGTH];
+	char			hash[(SHA256_DIGEST_LENGTH * 2) + 1];
+#endif
 	off_t			off;
 	void			*base;
-	struct mime_type	*mime;
 	struct buildopt		*bopt;
-	const char		*mime_type;
-	int			in, out, i, len;
-	u_int8_t		*d, digest[SHA256_DIGEST_LENGTH];
+	int			in, out;
 	char			*cpath, *ext, *opath, *p, *name;
-	char			hash[(SHA256_DIGEST_LENGTH * 2) + 1];
 
 	bopt = cli_buildopt_default();
 
@@ -985,7 +1120,7 @@ cli_build_asset(char *fpath, struct dirent *dp)
 
 	/* Replace dots, spaces, etc etc with underscores. */
 	for (p = name; *p != '\0'; p++) {
-		if (*p == '.' || isspace(*p) || *p == '-')
+		if (*p == '.' || isspace((unsigned char)*p) || *p == '-')
 			*p = '_';
 	}
 
@@ -1049,6 +1184,7 @@ cli_build_asset(char *fpath, struct dirent *dp)
 	 */
 	cli_file_writef(out, "0x00");
 
+#if !defined(KODEV_MINIMAL)
 	/* Calculate the SHA256 digest of the contents. */
 	(void)SHA256_Init(&sctx);
 	(void)SHA256_Update(&sctx, base, st.st_size);
@@ -1071,6 +1207,7 @@ cli_build_asset(char *fpath, struct dirent *dp)
 		mime_type = mime->type;
 	else
 		mime_type = "text/plain";
+#endif
 
 	/* Add the meta data. */
 	cli_file_writef(out, "};\n\n");
@@ -1080,6 +1217,7 @@ cli_build_asset(char *fpath, struct dirent *dp)
 	    "const time_t asset_mtime_%s_%s = %" PRI_TIME_T ";\n",
 	    name, ext, st.st_mtime);
 
+#if !defined(KODEV_MINIMAL)
 	if (bopt->flavor_nohttp == 0) {
 		cli_file_writef(out,
 		    "const char *asset_sha256_%s_%s = \"\\\"%s\\\"\";\n",
@@ -1087,6 +1225,7 @@ cli_build_asset(char *fpath, struct dirent *dp)
 		cli_file_writef(out, http_serveable_function,
 		    name, ext, name, ext, name, ext, name, ext, mime_type);
 	}
+#endif
 
 	/* Write the file symbols into assets.h so they can be used. */
 	cli_write_asset(name, ext, bopt);
@@ -1225,6 +1364,7 @@ cli_find_files(const char *path, void (*cb)(char *, struct dirent *))
 	closedir(d);
 }
 
+#if !defined(KODEV_MINIMAL)
 static void
 cli_generate_certs(void)
 {
@@ -1318,6 +1458,7 @@ cli_generate_certs(void)
 	EVP_PKEY_free(pkey);
 	X509_free(x509);
 }
+#endif
 
 static void
 cli_compile_source_file(void *arg)
@@ -1444,6 +1585,28 @@ cli_compile_kore(void *arg)
 	fatal("failed to start '%s': %s", args[0], errno_s);
 }
 
+static void
+cli_run_kore_python(void)
+{
+	char		*args[5], *cmd;
+	char		pwd[MAXPATHLEN];
+
+	(void)cli_vasprintf(&cmd, "%s/bin/kore", prefix);
+
+	if (getcwd(pwd, sizeof(pwd)) == NULL)
+		fatal("could not get cwd: %s", errno_s);
+
+	args[0] = cmd;
+	args[1] = "-frnc";
+	args[2] = "kore.conf";
+	args[3] = pwd;
+	args[4] = NULL;
+
+	execvp(args[0], args);
+	fatal("failed to start '%s': %s", args[0], errno_s);
+
+}
+
 static void
 cli_run_kore(void)
 {
