@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -45,19 +45,41 @@
 #include "curl.h"
 #endif
 
+#if defined(KORE_USE_ACME)
+#include "acme.h"
+#endif
+
+#if defined(__linux__)
+#include "seccomp.h"
+#endif
+
 /* XXX - This is becoming a clusterfuck. Fix it. */
 
-static int		configure_load(char *);
+static int	configure_load(char *);
+static void	configure_check_var(char **, const char *, const char *);
 
 #if defined(KORE_SINGLE_BINARY)
 static FILE		*config_file_write(void);
 extern u_int8_t		asset_builtin_kore_conf[];
 extern u_int32_t	asset_len_builtin_kore_conf;
+#elif defined(KORE_USE_PYTHON)
+static int		configure_file(char *);
 #endif
 
+#if defined(KORE_USE_ACME)
+static int		configure_acme(char *);
+static int		configure_acme_root(char *);
+static int		configure_acme_runas(char *);
+static int		configure_acme_email(char *);
+static int		configure_acme_provider(char *);
+#endif
+
+static int		configure_tls(char *);
+static int		configure_server(char *);
 static int		configure_include(char *);
 static int		configure_bind(char *);
 static int		configure_bind_unix(char *);
+static int		configure_attach(char *);
 static int		configure_domain(char *);
 static int		configure_root(char *);
 static int		configure_runas(char *);
@@ -74,7 +96,6 @@ static int		configure_socket_backlog(char *);
 static int		configure_add_pledge(char *);
 #endif
 
-#if !defined(KORE_NO_TLS)
 static int		configure_rand_file(char *);
 static int		configure_certfile(char *);
 static int		configure_certkey(char *);
@@ -85,14 +106,15 @@ static int		configure_keymgr_root(char *);
 static int		configure_keymgr_runas(char *);
 static int		configure_client_verify(char *);
 static int		configure_client_verify_depth(char *);
-#endif
 
 #if !defined(KORE_NO_HTTP)
+static int		configure_route(char *);
 static int		configure_filemap(char *);
-static int		configure_restrict(char *);
-static int		configure_handler(int, char *);
+static int		configure_return(char *);
+static int		configure_redirect(char *);
 static int		configure_static_handler(char *);
 static int		configure_dynamic_handler(char *);
+static int		configure_restrict(char *);
 static int		configure_accesslog(char *);
 static int		configure_http_header_max(char *);
 static int		configure_http_header_timeout(char *);
@@ -107,6 +129,8 @@ static int		configure_http_request_ms(char *);
 static int		configure_http_request_limit(char *);
 static int		configure_http_body_disk_offload(char *);
 static int		configure_http_body_disk_path(char *);
+static int		configure_http_server_version(char *);
+static int		configure_http_pretty_error(char *);
 static int		configure_validator(char *);
 static int		configure_params(char *);
 static int		configure_validate(char *);
@@ -129,6 +153,7 @@ static int		configure_task_threads(char *);
 #endif
 
 #if defined(KORE_USE_PYTHON)
+static int		configure_deployment(char *);
 static int		configure_python_path(char *);
 static int		configure_python_import(char *);
 #endif
@@ -138,21 +163,60 @@ static int		configure_curl_timeout(char *);
 static int		configure_curl_recv_max(char *);
 #endif
 
+#if defined(__linux__)
+static int		configure_seccomp_tracing(char *);
+#endif
+
 static struct {
 	const char		*name;
 	int			(*configure)(char *);
-} config_names[] = {
-	{ "include",			configure_include },
+} config_directives[] = {
+	{ "tls",			configure_tls },
+#if defined(KORE_USE_ACME)
+	{ "acme",			configure_acme },
+#endif
 	{ "bind",			configure_bind },
-	{ "bind_unix",			configure_bind_unix },
 	{ "load",			configure_load },
+	{ "domain",			configure_domain },
+	{ "server",			configure_server },
+	{ "attach",			configure_attach },
+	{ "certkey",			configure_certkey },
+	{ "certfile",			configure_certfile },
+	{ "include",			configure_include },
+	{ "unix",			configure_bind_unix },
+	{ "client_verify",		configure_client_verify },
+	{ "client_verify_depth",	configure_client_verify_depth },
 #if defined(KORE_USE_PYTHON)
 	{ "python_path",		configure_python_path },
 	{ "python_import",		configure_python_import },
 #endif
+#if !defined(KORE_NO_HTTP)
+	{ "route",			configure_route},
+	{ "filemap",			configure_filemap },
+	{ "redirect",			configure_redirect },
+	{ "return",			configure_return },
+	{ "static",			configure_static_handler },
+	{ "dynamic",			configure_dynamic_handler },
+	{ "accesslog",			configure_accesslog },
+	{ "restrict",			configure_restrict },
+	{ "validator",			configure_validator },
+	{ "params",			configure_params },
+	{ "validate",			configure_validate },
+	{ "authentication",		configure_authentication },
+	{ "authentication_uri",		configure_authentication_uri },
+	{ "authentication_type",	configure_authentication_type },
+	{ "authentication_value",	configure_authentication_value },
+	{ "authentication_validator",	configure_authentication_validator },
+#endif
+	{ NULL,				NULL },
+};
+
+static struct {
+	const char		*name;
+	int			(*configure)(char *);
+} config_settings[] = {
 	{ "root",			configure_root },
 	{ "chroot",			configure_root },
-	{ "domain",			configure_domain },
 	{ "runas",			configure_runas },
 	{ "workers",			configure_workers },
 	{ "worker_max_connections",	configure_max_connections },
@@ -162,29 +226,27 @@ static struct {
 	{ "worker_set_affinity",	configure_set_affinity },
 	{ "pidfile",			configure_pidfile },
 	{ "socket_backlog",		configure_socket_backlog },
-#if defined(KORE_USE_PLATFORM_PLEDGE)
-	{ "pledge",			configure_add_pledge },
-#endif
-#if !defined(KORE_NO_TLS)
 	{ "tls_version",		configure_tls_version },
 	{ "tls_cipher",			configure_tls_cipher },
 	{ "tls_dhparam",		configure_tls_dhparam },
 	{ "rand_file",			configure_rand_file },
 	{ "keymgr_runas",		configure_keymgr_runas },
 	{ "keymgr_root",		configure_keymgr_root },
-	{ "certfile",			configure_certfile },
-	{ "certkey",			configure_certkey },
-	{ "client_verify",		configure_client_verify },
-	{ "client_verify_depth",	configure_client_verify_depth },
+#if defined(KORE_USE_ACME)
+	{ "acme_runas",			configure_acme_runas },
+	{ "acme_root",			configure_acme_root },
+	{ "acme_email",			configure_acme_email },
+	{ "acme_provider",		configure_acme_provider },
+#endif
+#if defined(KORE_USE_PLATFORM_PLEDGE)
+	{ "pledge",			configure_add_pledge },
+#endif
+#if defined(__linux__)
+	{ "seccomp_tracing",		configure_seccomp_tracing },
 #endif
 #if !defined(KORE_NO_HTTP)
-	{ "filemap",			configure_filemap },
 	{ "filemap_ext",		configure_filemap_ext },
 	{ "filemap_index",		configure_filemap_index },
-	{ "static",			configure_static_handler },
-	{ "dynamic",			configure_dynamic_handler },
-	{ "accesslog",			configure_accesslog },
-	{ "restrict",			configure_restrict },
 	{ "http_media_type",		configure_http_media_type },
 	{ "http_header_max",		configure_http_header_max },
 	{ "http_header_timeout",	configure_http_header_timeout },
@@ -196,17 +258,14 @@ static struct {
 	{ "http_request_limit",		configure_http_request_limit },
 	{ "http_body_disk_offload",	configure_http_body_disk_offload },
 	{ "http_body_disk_path",	configure_http_body_disk_path },
-	{ "validator",			configure_validator },
-	{ "params",			configure_params },
-	{ "validate",			configure_validate },
-	{ "authentication",		configure_authentication },
-	{ "authentication_uri",		configure_authentication_uri },
-	{ "authentication_type",	configure_authentication_type },
-	{ "authentication_value",	configure_authentication_value },
-	{ "authentication_validator",	configure_authentication_validator },
+	{ "http_server_version",	configure_http_server_version },
+	{ "http_pretty_error",		configure_http_pretty_error },
 	{ "websocket_maxframe",		configure_websocket_maxframe },
 	{ "websocket_timeout",		configure_websocket_timeout },
 #endif
+#if defined(KORE_USE_PYTHON)
+	{ "deployment",			configure_deployment },
+#endif
 #if defined(KORE_USE_PGSQL)
 	{ "pgsql_conn_max",		configure_pgsql_conn_max },
 	{ "pgsql_queue_limit",		configure_pgsql_queue_limit },
@@ -217,10 +276,15 @@ static struct {
 #if defined(KORE_USE_CURL)
 	{ "curl_timeout",		configure_curl_timeout },
 	{ "curl_recv_max",		configure_curl_recv_max },
+#endif
+#if !defined(KORE_SINGLE_BINARY) && defined(KORE_USE_PYTHON)
+	{ "file",			configure_file },
 #endif
 	{ NULL,				NULL },
 };
 
+static int				finalized = 0;
+
 #if !defined(KORE_SINGLE_BINARY)
 char					*config_file = NULL;
 #endif
@@ -234,6 +298,7 @@ static struct kore_module_handle	*current_handler = NULL;
 
 extern const char			*__progname;
 static struct kore_domain		*current_domain = NULL;
+static struct kore_server		*current_server = NULL;
 
 void
 kore_parse_config(void)
@@ -241,15 +306,26 @@ kore_parse_config(void)
 	FILE		*fp;
 	char		path[PATH_MAX];
 
+	if (finalized)
+		return;
+
+	fp = NULL;
+
 #if !defined(KORE_SINGLE_BINARY)
-	if ((fp = fopen(config_file, "r")) == NULL)
-		fatal("configuration given cannot be opened: %s", config_file);
+	if (config_file != NULL) {
+		if ((fp = fopen(config_file, "r")) == NULL) {
+			fatal("configuration given cannot be opened: %s",
+			    config_file);
+		}
+	}
 #else
 	fp = config_file_write();
 #endif
 
-	kore_parse_config_file(fp);
-	(void)fclose(fp);
+	if (fp != NULL) {
+		kore_parse_config_file(fp);
+		(void)fclose(fp);
+	}
 
 	if (!kore_module_loaded())
 		fatal("no application module was loaded");
@@ -281,29 +357,26 @@ kore_parse_config(void)
 		if (!kore_quiet)
 			kore_log(LOG_WARNING, "privsep: will not change user");
 	} else {
-#if !defined(KORE_NO_TLS)
-		if (keymgr_runas_user == NULL) {
-			if (!kore_quiet) {
-				kore_log(LOG_NOTICE, "privsep: no keymgr_runas "
-				    "set, using 'runas` user");
-			}
-			keymgr_runas_user = kore_strdup(kore_runas_user);
-		}
+		configure_check_var(&keymgr_runas_user, kore_runas_user,
+			"privsep: no keymgr_runas set, using 'runas` user");
+#if defined(KORE_USE_ACME)
+		configure_check_var(&acme_runas_user, kore_runas_user,
+			"privsep: no acme_runas set, using 'runas` user");
 #endif
 	}
 
-#if !defined(KORE_NO_TLS)
-	if (keymgr_root_path == NULL) {
-		if (!kore_quiet) {
-			kore_log(LOG_NOTICE, "privsep: no keymgr_root set, "
-			    "using 'root` directory");
-		}
-		keymgr_root_path = kore_strdup(kore_root_path);
-	}
+	configure_check_var(&keymgr_root_path, kore_root_path,
+		"privsep: no keymgr_root set, using 'root` directory");
+
+#if defined(KORE_USE_ACME)
+	configure_check_var(&acme_root_path, kore_root_path,
+		"privsep: no acme_root set, using 'root` directory");
 #endif
 
 	if (skip_chroot && !kore_quiet)
 		kore_log(LOG_WARNING, "privsep: will not chroot");
+
+	finalized = 1;
 }
 
 void
@@ -319,6 +392,13 @@ kore_parse_config_file(FILE *fp)
 			continue;
 		}
 
+		if (!strcmp(p, "}") && current_server != NULL) {
+			lineno++;
+			kore_server_finalize(current_server);
+			current_server = NULL;
+			continue;
+		}
+
 #if !defined(KORE_NO_HTTP)
 		if (!strcmp(p, "}") && current_handler != NULL) {
 			lineno++;
@@ -341,13 +421,26 @@ kore_parse_config_file(FILE *fp)
 #endif
 
 		if (!strcmp(p, "}") && current_domain != NULL) {
-#if !defined(KORE_NO_TLS)
-			if (current_domain->certfile == NULL ||
-			    current_domain->certkey == NULL) {
-				fatal("incomplete TLS setup for '%s'",
+			if (current_domain->server == NULL) {
+				fatal("domain '%s' not attached to server",
 				    current_domain->domain);
 			}
+
+			if (current_domain->server->tls == 1) {
+#if defined(KORE_USE_ACME)
+				if (current_domain->acme) {
+					lineno++;
+					current_domain = NULL;
+					continue;
+				}
 #endif
+				if (current_domain->certfile == NULL ||
+				    current_domain->certkey == NULL) {
+					fatal("incomplete TLS setup for '%s'",
+					    current_domain->domain);
+				}
+			}
+
 			current_domain = NULL;
 		}
 
@@ -371,21 +464,68 @@ kore_parse_config_file(FILE *fp)
 			continue;
 		}
 
-		for (i = 0; config_names[i].name != NULL; i++) {
-			if (!strcmp(config_names[i].name, p)) {
-				if (config_names[i].configure(t))
+		for (i = 0; config_directives[i].name != NULL; i++) {
+			if (!strcmp(config_directives[i].name, p)) {
+				if (config_directives[i].configure(t))
 					break;
 				fatal("configuration error on line %d", lineno);
 				/* NOTREACHED */
 			}
 		}
 
-		if (config_names[i].name == NULL)
+		if (config_directives[i].name != NULL) {
+			lineno++;
+			continue;
+		}
+
+		for (i = 0; config_settings[i].name != NULL; i++) {
+			if (!strcmp(config_settings[i].name, p)) {
+				if (config_settings[i].configure(t))
+					break;
+				fatal("configuration error on line %d", lineno);
+				/* NOTREACHED */
+			}
+		}
+
+		if (config_settings[i].name == NULL)
 			printf("ignoring \"%s\" on line %d\n", p, lineno);
+
 		lineno++;
 	}
 }
 
+#if defined(KORE_USE_PYTHON)
+int
+kore_configure_setting(const char *name, char *value)
+{
+	int	i;
+
+	if (finalized)
+		return (KORE_RESULT_ERROR);
+
+	for (i = 0; config_settings[i].name != NULL; i++) {
+		if (!strcmp(config_settings[i].name, name)) {
+			if (config_settings[i].configure(value))
+				return (KORE_RESULT_OK);
+			fatal("bad value '%s' for '%s'", value, name);
+		}
+	}
+
+	kore_log(LOG_NOTICE, "ignoring unknown kore.config.%s setting", name);
+	return (KORE_RESULT_OK);
+}
+#endif
+
+static void
+configure_check_var(char **var, const char *other, const char *logmsg)
+{
+	if (*var == NULL) {
+		if (!kore_quiet)
+			kore_log(LOG_NOTICE, "%s", logmsg);
+		*var = kore_strdup(other);
+	}
+}
+
 static int
 configure_include(char *path)
 {
@@ -400,16 +540,146 @@ configure_include(char *path)
 	return (KORE_RESULT_OK);
 }
 
+static int
+configure_server(char *options)
+{
+	struct kore_server	*srv;
+	char			*argv[3];
+
+	if (current_server != NULL) {
+		printf("nested server contexts are not allowed\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	kore_split_string(options, " ", argv, 3);
+
+	if (argv[0] == NULL || argv[1] == NULL) {
+		printf("invalid server context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (strcmp(argv[1], "{")) {
+		printf("server context not opened correctly\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	if ((srv = kore_server_lookup(argv[0])) != NULL) {
+		printf("a server with name '%s' already exists\n", srv->name);
+		return (KORE_RESULT_ERROR);
+	}
+
+	current_server = kore_server_create(argv[0]);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_tls(char *yesno)
+{
+	if (current_server == NULL) {
+		printf("bind directive not inside a server context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (!strcmp(yesno, "no")) {
+		current_server->tls = 0;
+	} else if (!strcmp(yesno, "yes")) {
+		current_server->tls = 1;
+	} else {
+		printf("invalid '%s' for yes|no tls option\n", yesno);
+		return (KORE_RESULT_ERROR);
+	}
+
+	return (KORE_RESULT_OK);
+}
+
+#if defined(KORE_USE_ACME)
+static int
+configure_acme(char *yesno)
+{
+	if (current_domain == NULL) {
+		printf("acme directive not inside a domain context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (strchr(current_domain->domain, '*')) {
+		printf("wildcards not supported due to lack of dns-01\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (!strcmp(yesno, "no")) {
+		current_domain->acme = 0;
+	} else if (!strcmp(yesno, "yes")) {
+		current_domain->acme = 1;
+
+		/* Override keyfile and certfile locations. */
+		kore_free(current_domain->certkey);
+		kore_free(current_domain->certfile);
+
+		kore_acme_get_paths(current_domain->domain,
+		    &current_domain->certkey, &current_domain->certfile);
+
+	} else {
+		printf("invalid '%s' for yes|no acme option\n", yesno);
+		return (KORE_RESULT_ERROR);
+	}
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_acme_runas(char *user)
+{
+	kore_free(acme_runas_user);
+	acme_runas_user = kore_strdup(user);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_acme_root(char *root)
+{
+	kore_free(acme_root_path);
+	acme_root_path = kore_strdup(root);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_acme_email(char *email)
+{
+	kore_free(acme_email);
+	acme_email = kore_strdup(email);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_acme_provider(char *provider)
+{
+	kore_free(acme_provider);
+	acme_provider = kore_strdup(provider);
+
+	return (KORE_RESULT_OK);
+}
+
+#endif
+
 static int
 configure_bind(char *options)
 {
 	char		*argv[4];
 
+	if (current_server == NULL) {
+		printf("bind directive not inside a server context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
 	kore_split_string(options, " ", argv, 4);
 	if (argv[0] == NULL || argv[1] == NULL)
 		return (KORE_RESULT_ERROR);
 
-	return (kore_server_bind(argv[0], argv[1], argv[2]));
+	return (kore_server_bind(current_server, argv[0], argv[1], argv[2]));
 }
 
 static int
@@ -417,11 +687,16 @@ configure_bind_unix(char *options)
 {
 	char		*argv[3];
 
+	if (current_server == NULL) {
+		printf("bind_unix directive not inside a server context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
 	kore_split_string(options, " ", argv, 3);
 	if (argv[0] == NULL)
 		return (KORE_RESULT_ERROR);
 
-	return (kore_server_bind_unix(argv[0], argv[1]));
+	return (kore_server_bind_unix(current_server, argv[0], argv[1]));
 }
 
 static int
@@ -476,9 +751,18 @@ config_file_write(void)
 
 	return (fp);
 }
+#elif defined(KORE_USE_PYTHON)
+static int
+configure_file(char *file)
+{
+	free(config_file);
+	if ((config_file = strdup(file)) == NULL)
+		fatal("strdup");
+
+	return (KORE_RESULT_OK);
+}
 #endif
 
-#if !defined(KORE_NO_TLS)
 static int
 configure_tls_version(char *version)
 {
@@ -603,12 +887,7 @@ configure_certfile(char *path)
 		return (KORE_RESULT_ERROR);
 	}
 
-	if (current_domain->certfile != NULL) {
-		printf("certfile specified twice for %s\n",
-		    current_domain->domain);
-		return (KORE_RESULT_ERROR);
-	}
-
+	kore_free(current_domain->certfile);
 	current_domain->certfile = kore_strdup(path);
 	return (KORE_RESULT_OK);
 }
@@ -621,12 +900,7 @@ configure_certkey(char *path)
 		return (KORE_RESULT_ERROR);
 	}
 
-	if (current_domain->certkey != NULL) {
-		printf("certkey specified twice for %s\n",
-		    current_domain->domain);
-		return (KORE_RESULT_ERROR);
-	}
-
+	kore_free(current_domain->certkey);
 	current_domain->certkey = kore_strdup(path);
 	return (KORE_RESULT_OK);
 }
@@ -651,8 +925,6 @@ configure_keymgr_root(char *root)
 	return (KORE_RESULT_OK);
 }
 
-#endif /* !KORE_NO_TLS */
-
 static int
 configure_domain(char *options)
 {
@@ -665,6 +937,11 @@ configure_domain(char *options)
 
 	kore_split_string(options, " ", argv, 3);
 
+	if (argv[0] == NULL || argv[1] == NULL) {
+		printf("invalid domain context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
 	if (strcmp(argv[1], "{")) {
 		printf("domain context not opened correctly\n");
 		return (KORE_RESULT_ERROR);
@@ -675,12 +952,38 @@ configure_domain(char *options)
 		return (KORE_RESULT_ERROR);
 	}
 
-	if (!kore_domain_new(argv[0])) {
-		printf("could not create new domain %s\n", argv[0]);
+	current_domain = kore_domain_new(argv[0]);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_attach(char *name)
+{
+	struct kore_server	*srv;
+
+	if (current_domain == NULL) {
+		printf("attach not specified in domain context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (current_domain->server != NULL) {
+		printf("domain '%s' already attached to server\n",
+		    current_domain->domain);
+		return (KORE_RESULT_ERROR);
+	}
+
+	if ((srv = kore_server_lookup(name)) == NULL) {
+		printf("server '%s' does not exist\n", name);
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (!kore_domain_attach(current_domain, srv)) {
+		printf("failed to attach '%s' to '%s'\n",
+		    current_domain->domain, name);
 		return (KORE_RESULT_ERROR);
 	}
 
-	current_domain = kore_domain_lookup(argv[0]);
 	return (KORE_RESULT_OK);
 }
 
@@ -688,35 +991,105 @@ configure_domain(char *options)
 static int
 configure_static_handler(char *options)
 {
-	return (configure_handler(HANDLER_TYPE_STATIC, options));
+	kore_log(LOG_NOTICE, "static keyword removed, use route instead");
+	return (KORE_RESULT_ERROR);
 }
 
 static int
 configure_dynamic_handler(char *options)
 {
-	return (configure_handler(HANDLER_TYPE_DYNAMIC, options));
+	kore_log(LOG_NOTICE, "dynamic keyword removed, use route instead");
+	return (KORE_RESULT_ERROR);
 }
 
 static int
-configure_handler(int type, char *options)
+configure_route(char *options)
 {
+	int		type;
 	char		*argv[4];
 
 	if (current_domain == NULL) {
-		printf("page handler not specified in domain context\n");
+		printf("route not specified in domain context\n");
 		return (KORE_RESULT_ERROR);
 	}
 
 	kore_split_string(options, " ", argv, 4);
 
 	if (argv[0] == NULL || argv[1] == NULL) {
-		printf("missing parameters for page handler\n");
+		printf("missing parameters for route \n");
 		return (KORE_RESULT_ERROR);
 	}
 
-	if (!kore_module_handler_new(argv[0],
-	    current_domain->domain, argv[1], argv[2], type)) {
-		printf("cannot create handler for %s\n", argv[0]);
+	if (*argv[0] == '/')
+		type = HANDLER_TYPE_STATIC;
+	else
+		type = HANDLER_TYPE_DYNAMIC;
+
+	if (!kore_module_handler_new(current_domain,
+	    argv[0], argv[1], argv[2], type)) {
+		printf("cannot create route for %s\n", argv[0]);
+		return (KORE_RESULT_ERROR);
+	}
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_return(char *options)
+{
+	char		*argv[3];
+	int		elm, status, err;
+
+	if (current_domain == NULL) {
+		printf("return outside of domain context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	elm = kore_split_string(options, " ", argv, 3);
+	if (elm != 2) {
+		printf("missing parameters for return\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	status = kore_strtonum(argv[1], 10, 400, 600, &err);
+	if (err != KORE_RESULT_OK) {
+		printf("invalid status code on return (%s)\n", argv[1]);
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (!http_redirect_add(current_domain, argv[0], status, NULL)) {
+		printf("invalid regex on return path\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_redirect(char *options)
+{
+	char		*argv[4];
+	int		elm, status, err;
+
+	if (current_domain == NULL) {
+		printf("redirect outside of domain context\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	elm = kore_split_string(options, " ", argv, 4);
+	if (elm != 3) {
+		printf("missing parameters for redirect\n");
+		return (KORE_RESULT_ERROR);
+	}
+
+	status = kore_strtonum(argv[1], 10, 300, 399, &err);
+	if (err != KORE_RESULT_OK) {
+		printf("invalid status code on redirect (%s)\n", argv[1]);
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (!http_redirect_add(current_domain, argv[0], status, argv[2])) {
+		printf("invalid regex on redirect path\n");
 		return (KORE_RESULT_ERROR);
 	}
 
@@ -955,6 +1328,30 @@ configure_http_body_disk_path(char *path)
 	return (KORE_RESULT_OK);
 }
 
+static int
+configure_http_server_version(char *version)
+{
+	http_server_version(version);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+configure_http_pretty_error(char *yesno)
+{
+	if (!strcmp(yesno, "no")) {
+		http_pretty_error = 0;
+	} else if (!strcmp(yesno, "yes")) {
+		http_pretty_error = 1;
+	} else {
+		printf("invalid '%s' for yes|no http_pretty_error option\n",
+		    yesno);
+		return (KORE_RESULT_ERROR);
+	}
+
+	return (KORE_RESULT_OK);
+}
+
 static int
 configure_http_hsts_enable(char *option)
 {
@@ -1306,7 +1703,7 @@ configure_workers(char *option)
 {
 	int		err;
 
-	worker_count = kore_strtonum(option, 10, 1, 255, &err);
+	worker_count = kore_strtonum(option, 10, 1, KORE_WORKER_MAX, &err);
 	if (err != KORE_RESULT_OK) {
 		printf("%s is not a valid worker number\n", option);
 		return (KORE_RESULT_ERROR);
@@ -1457,6 +1854,29 @@ configure_task_threads(char *option)
 #endif
 
 #if defined(KORE_USE_PYTHON)
+static int
+configure_deployment(char *value)
+{
+	if (!strcmp(value, "dev") || !strcmp(value, "development")) {
+		foreground = 1;
+		skip_runas = 1;
+		skip_chroot = 1;
+	} else if (!strcmp(value, "production")) {
+		foreground = 0;
+		skip_runas = 0;
+		skip_chroot = 0;
+	} else {
+		kore_log(LOG_NOTICE,
+		    "pyko.config.deployment: bad value '%s'", value);
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (!kore_quiet)
+		kore_log(LOG_NOTICE, "deployment set to %s", value);
+
+	return (KORE_RESULT_OK);
+}
+
 static int
 configure_python_path(char *path)
 {
@@ -1518,3 +1938,21 @@ configure_curl_timeout(char *option)
 	return (KORE_RESULT_OK);
 }
 #endif
+
+#if defined(__linux__)
+static int
+configure_seccomp_tracing(char *opt)
+{
+	if (!strcmp(opt, "yes")) {
+		kore_seccomp_tracing = 1;
+	} else if (!strcmp(opt, "no")) {
+		kore_seccomp_tracing = 0;
+	} else {
+		printf("bad seccomp_tracing value: %s (expected yes|no)\n",
+		    opt);
+		return (KORE_RESULT_ERROR);
+	}
+
+	return (KORE_RESULT_OK);
+}
+#endif
