@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -23,14 +23,12 @@
 #include <sys/param.h>
 #include <sys/types.h>
 
-#if !defined(KORE_NO_TLS)
 #include <openssl/x509.h>
 #include <openssl/bio.h>
 #include <openssl/evp.h>
 #include <openssl/ec.h>
 #include <openssl/ecdsa.h>
 #include <poll.h>
-#endif
 
 #include <fnmatch.h>
 
@@ -40,22 +38,21 @@
 #include "http.h"
 #endif
 
+#if defined(KORE_USE_ACME)
+#include "acme.h"
+#endif
+
 #define KORE_DOMAIN_CACHE	16
 #define SSL_SESSION_ID		"kore_ssl_sessionid"
 
-struct kore_domain_h		domains;
 struct kore_domain		*primary_dom = NULL;
 
-#if !defined(KORE_NO_TLS)
 static u_int8_t			keymgr_buf[2048];
 static size_t			keymgr_buflen = 0;
 static int			keymgr_response = 0;
 DH				*tls_dhparam = NULL;
 int				tls_version = KORE_TLS_VERSION_BOTH;
-#endif
 
-#if !defined(KORE_NO_TLS)
-static BIO	*domain_bio_mem(const void *, size_t);
 static int	domain_x509_verify(int, X509_STORE_CTX *);
 static X509	*domain_load_certificate_chain(SSL_CTX *, const void *, size_t);
 
@@ -71,11 +68,10 @@ static int	keymgr_rsa_privenc(int, const unsigned char *,
 static ECDSA_SIG	*keymgr_ecdsa_sign(const unsigned char *, int,
 			    const BIGNUM *, const BIGNUM *, EC_KEY *);
 
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 static RSA_METHOD	*keymgr_rsa_meth = NULL;
 static EC_KEY_METHOD	*keymgr_ec_meth = NULL;
 #else
-#if !defined(LIBRESSL_VERSION_TEXT)
 /*
  * Run own ecdsa_method data structure as OpenSSL has this in ecs_locl.h
  * and does not export this on systems.
@@ -93,6 +89,7 @@ struct ecdsa_method {
 };
 #endif
 
+#if !defined(KORE_OPENSSL_NEWER_API)
 static ECDSA_METHOD	keymgr_ecdsa = {
 	"kore ECDSA keymgr method",
 	keymgr_ecdsa_sign,
@@ -118,9 +115,7 @@ static RSA_METHOD	keymgr_rsa = {
 	NULL,
 	NULL
 };
-
-#endif /* OPENSSL_VERSION_NUMBER */
-#endif /* KORE_NO_TLS */
+#endif
 
 static u_int16_t		domain_id = 0;
 static struct kore_domain	*cached[KORE_DOMAIN_CACHE];
@@ -130,13 +125,10 @@ kore_domain_init(void)
 {
 	int		i;
 
-	TAILQ_INIT(&domains);
-
 	for (i = 0; i < KORE_DOMAIN_CACHE; i++)
 		cached[i] = NULL;
 
-#if !defined(KORE_NO_TLS)
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	if (keymgr_rsa_meth == NULL) {
 		if ((keymgr_rsa_meth = RSA_meth_new("kore RSA keymgr method",
 		    RSA_METHOD_FLAG_NO_CHECK)) == NULL)
@@ -162,22 +154,12 @@ kore_domain_init(void)
 		    OPENSSL_VERSION_TEXT);
 	}
 #endif
-
-#endif
 }
 
 void
 kore_domain_cleanup(void)
 {
-	struct kore_domain *dom;
-
-	while ((dom = TAILQ_FIRST(&domains)) != NULL) {
-		TAILQ_REMOVE(&domains, dom, list);
-		kore_domain_free(dom);
-	}
-
-#if !defined(KORE_NO_TLS)
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	if (keymgr_rsa_meth != NULL) {
 		RSA_meth_free(keymgr_rsa_meth);
 		keymgr_rsa_meth = NULL;
@@ -188,37 +170,25 @@ kore_domain_cleanup(void)
 		keymgr_ec_meth = NULL;
 	}
 #endif
-#endif
 }
 
-int
-kore_domain_new(char *domain)
+struct kore_domain *
+kore_domain_new(const char *domain)
 {
 	struct kore_domain	*dom;
 
-	if (kore_domain_lookup(domain) != NULL)
-		return (KORE_RESULT_ERROR);
-
 	kore_debug("kore_domain_new(%s)", domain);
 
-	dom = kore_malloc(sizeof(*dom));
+	dom = kore_calloc(1, sizeof(*dom));
 	dom->id = domain_id++;
-	dom->logbuf = NULL;
 	dom->accesslog = -1;
-	dom->logwarn = 0;
-	dom->logerr = 0;
-#if !defined(KORE_NO_TLS)
-	dom->cafile = NULL;
-	dom->certkey = NULL;
-	dom->ssl_ctx = NULL;
-	dom->certfile = NULL;
-	dom->crlfile = NULL;
+
 	dom->x509_verify_depth = 1;
-#endif
 	dom->domain = kore_strdup(domain);
 
 #if !defined(KORE_NO_HTTP)
 	TAILQ_INIT(&(dom->handlers));
+	TAILQ_INIT(&(dom->redirects));
 #endif
 
 	if (dom->id < KORE_DOMAIN_CACHE) {
@@ -227,11 +197,32 @@ kore_domain_new(char *domain)
 		cached[dom->id] = dom;
 	}
 
-	TAILQ_INSERT_TAIL(&domains, dom, list);
-
 	if (primary_dom == NULL)
 		primary_dom = dom;
 
+	return (dom);
+}
+
+int
+kore_domain_attach(struct kore_domain *dom, struct kore_server *server)
+{
+	struct kore_domain	*d;
+
+	if (dom->server != NULL)
+		return (KORE_RESULT_ERROR);
+
+	TAILQ_FOREACH(d, &server->domains, list) {
+		if (!strcmp(d->domain, dom->domain))
+			return (KORE_RESULT_ERROR);
+	}
+
+	dom->server = server;
+	TAILQ_INSERT_TAIL(&server->domains, dom, list);
+
+	/* The primary domain should be attached to a TLS context. */
+	if (server->tls == 0 && dom == primary_dom)
+		primary_dom = NULL;
+
 	return (KORE_RESULT_OK);
 }
 
@@ -239,7 +230,8 @@ void
 kore_domain_free(struct kore_domain *dom)
 {
 #if !defined(KORE_NO_HTTP)
-	struct kore_module_handle *hdlr;
+	struct http_redirect		*rdr;
+	struct kore_module_handle	*hdlr;
 #endif
 	if (dom == NULL)
 		return;
@@ -247,12 +239,11 @@ kore_domain_free(struct kore_domain *dom)
 	if (primary_dom == dom)
 		primary_dom = NULL;
 
-	TAILQ_REMOVE(&domains, dom, list);
+	TAILQ_REMOVE(&dom->server->domains, dom, list);
 
 	if (dom->domain != NULL)
 		kore_free(dom->domain);
 
-#if !defined(KORE_NO_TLS)
 	if (dom->ssl_ctx != NULL)
 		SSL_CTX_free(dom->ssl_ctx);
 	if (dom->cafile != NULL)
@@ -263,7 +254,6 @@ kore_domain_free(struct kore_domain *dom)
 		kore_free(dom->certfile);
 	if (dom->crlfile != NULL)
 		kore_free(dom->crlfile);
-#endif
 
 #if !defined(KORE_NO_HTTP)
 	/* Drop all handlers associated with this domain */
@@ -271,21 +261,29 @@ kore_domain_free(struct kore_domain *dom)
 		TAILQ_REMOVE(&(dom->handlers), hdlr, list);
 		kore_module_handler_free(hdlr);
 	}
+
+	while ((rdr = TAILQ_FIRST(&(dom->redirects))) != NULL) {
+		TAILQ_REMOVE(&(dom->redirects), rdr, list);
+		regfree(&rdr->rctx);
+		kore_free(rdr->target);
+		kore_free(rdr);
+	}
 #endif
 	kore_free(dom);
 }
 
-#if !defined(KORE_NO_TLS)
 void
-kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
+kore_domain_tlsinit(struct kore_domain *dom, int type,
+    const void *data, size_t datalen)
 {
+	const u_int8_t		*ptr;
 	RSA			*rsa;
 	X509			*x509;
 	EVP_PKEY		*pkey;
 	STACK_OF(X509_NAME)	*certs;
 	EC_KEY			*eckey;
 	const SSL_METHOD	*method;
-#if defined(LIBRESSL_VERSION_TEXT) || OPENSSL_VERSION_NUMBER < 0x10100000L
+#if !defined(KORE_OPENSSL_NEWER_API)
 	EC_KEY			*ecdh;
 #endif
 
@@ -294,7 +292,7 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	if (dom->ssl_ctx != NULL)
 		SSL_CTX_free(dom->ssl_ctx);
 
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	if ((method = TLS_method()) == NULL)
 		fatalx("TLS_method(): %s", ssl_errno_s);
 #else
@@ -313,7 +311,7 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	if ((dom->ssl_ctx = SSL_CTX_new(method)) == NULL)
 		fatalx("SSL_ctx_new(): %s", ssl_errno_s);
 
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	if (!SSL_CTX_set_min_proto_version(dom->ssl_ctx, TLS1_2_VERSION))
 		fatalx("SSL_CTX_set_min_proto_version: %s", ssl_errno_s);
 
@@ -350,7 +348,30 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	}
 #endif
 
-	x509 = domain_load_certificate_chain(dom->ssl_ctx, pem, pemlen);
+	switch (type) {
+	case KORE_PEM_CERT_CHAIN:
+		x509 = domain_load_certificate_chain(dom->ssl_ctx,
+		    data, datalen);
+		break;
+	case KORE_DER_CERT_DATA:
+		ptr = data;
+		if ((x509 = d2i_X509(NULL, &ptr, datalen)) == NULL)
+			fatalx("d2i_X509: %s", ssl_errno_s);
+		if (SSL_CTX_use_certificate(dom->ssl_ctx, x509) == 0)
+			fatalx("SSL_CTX_use_certificate: %s", ssl_errno_s);
+		break;
+	default:
+		fatalx("%s: unknown type %d", __func__, type);
+	}
+
+	if (x509 == NULL) {
+		kore_log(LOG_NOTICE, "failed to load certificate for '%s': %s",
+		    dom->domain, ssl_errno_s);
+		SSL_CTX_free(dom->ssl_ctx);
+		dom->ssl_ctx = NULL;
+		return;
+	}
+
 	if ((pkey = X509_get_pubkey(x509)) == NULL)
 		fatalx("certificate has no public key");
 
@@ -359,7 +380,7 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 		if ((rsa = EVP_PKEY_get1_RSA(pkey)) == NULL)
 			fatalx("no RSA public key present");
 		RSA_set_app_data(rsa, dom);
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 		RSA_set_method(rsa, keymgr_rsa_meth);
 #else
 		RSA_set_method(rsa, &keymgr_rsa);
@@ -368,7 +389,7 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	case EVP_PKEY_EC:
 		if ((eckey = EVP_PKEY_get1_EC_KEY(pkey)) == NULL)
 			fatalx("no EC public key present");
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 		EC_KEY_set_ex_data(eckey, 0, dom);
 		EC_KEY_set_method(eckey, keymgr_ec_meth);
 #else
@@ -383,8 +404,10 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	if (!SSL_CTX_use_PrivateKey(dom->ssl_ctx, pkey))
 		fatalx("SSL_CTX_use_PrivateKey(): %s", ssl_errno_s);
 
-	if (!SSL_CTX_check_private_key(dom->ssl_ctx))
-		fatalx("Public/Private key for %s do not match", dom->domain);
+	if (!SSL_CTX_check_private_key(dom->ssl_ctx)) {
+		fatalx("Public/Private key for %s do not match (%s)",
+		    dom->domain, ssl_errno_s);
+	}
 
 	if (tls_dhparam == NULL)
 		fatalx("No DH parameters given");
@@ -392,7 +415,7 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	SSL_CTX_set_tmp_dh(dom->ssl_ctx, tls_dhparam);
 	SSL_CTX_set_options(dom->ssl_ctx, SSL_OP_SINGLE_DH_USE);
 
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	if (!SSL_CTX_set_ecdh_auto(dom->ssl_ctx, 1))
 		fatalx("SSL_CTX_set_ecdh_auto: %s", ssl_errno_s);
 #else
@@ -436,6 +459,10 @@ kore_domain_tlsinit(struct kore_domain *dom, const void *pem, size_t pemlen)
 	SSL_CTX_set_info_callback(dom->ssl_ctx, kore_tls_info_callback);
 	SSL_CTX_set_tlsext_servername_callback(dom->ssl_ctx, kore_tls_sni_cb);
 
+#if defined(KORE_USE_ACME)
+	SSL_CTX_set_alpn_select_cb(dom->ssl_ctx, kore_acme_tls_alpn, dom);
+#endif
+
 	X509_free(x509);
 }
 
@@ -448,7 +475,7 @@ kore_domain_crl_add(struct kore_domain *dom, const void *pem, size_t pemlen)
 	X509_STORE		*store;
 
 	ERR_clear_error();
-	in = domain_bio_mem(pem, pemlen);
+	in = BIO_new_mem_buf(pem, pemlen);
 
 	if ((store = SSL_CTX_get_cert_store(dom->ssl_ctx)) == NULL) {
 		BIO_free(in);
@@ -489,24 +516,26 @@ kore_domain_crl_add(struct kore_domain *dom, const void *pem, size_t pemlen)
 	X509_STORE_set_flags(store,
 	    X509_V_FLAG_CRL_CHECK | X509_V_FLAG_CRL_CHECK_ALL);
 }
-#endif
 
 void
 kore_domain_callback(void (*cb)(struct kore_domain *))
 {
+	struct kore_server	*srv;
 	struct kore_domain	*dom;
 
-	TAILQ_FOREACH(dom, &domains, list) {
-		cb(dom);
+	LIST_FOREACH(srv, &kore_servers, list) {
+		TAILQ_FOREACH(dom, &srv->domains, list) {
+			cb(dom);
+		}
 	}
 }
 
 struct kore_domain *
-kore_domain_lookup(const char *domain)
+kore_domain_lookup(struct kore_server *srv, const char *domain)
 {
 	struct kore_domain	*dom;
 
-	TAILQ_FOREACH(dom, &domains, list) {
+	TAILQ_FOREACH(dom, &srv->domains, list) {
 		if (!strcmp(dom->domain, domain))
 			return (dom);
 		if (!fnmatch(dom->domain, domain, FNM_CASEFOLD))
@@ -519,14 +548,17 @@ kore_domain_lookup(const char *domain)
 struct kore_domain *
 kore_domain_byid(u_int16_t id)
 {
+	struct kore_server	*srv;
 	struct kore_domain	*dom;
 
 	if (id < KORE_DOMAIN_CACHE)
 		return (cached[id]);
 
-	TAILQ_FOREACH(dom, &domains, list) {
-		if (dom->id == id)
-			return (dom);
+	LIST_FOREACH(srv, &kore_servers, list) {
+		TAILQ_FOREACH(dom, &srv->domains, list) {
+			if (dom->id == id)
+				return (dom);
+		}
 	}
 
 	return (NULL);
@@ -539,15 +571,21 @@ kore_domain_byid(u_int16_t id)
 void
 kore_domain_closelogs(void)
 {
+	struct kore_server	*srv;
 	struct kore_domain	*dom;
 
-	TAILQ_FOREACH(dom, &domains, list) {
-		if (dom->accesslog != -1) {
-			(void)close(dom->accesslog);
-			/* turn into flag to indicate accesslogs are active. */
-			dom->accesslog = 1;
-		} else {
-			dom->accesslog = 0;
+	LIST_FOREACH(srv, &kore_servers, list) {
+		TAILQ_FOREACH(dom, &srv->domains, list) {
+			if (dom->accesslog != -1) {
+				(void)close(dom->accesslog);
+				/*
+				 * Turn into flag to indicate accesslogs
+				 * are active.
+				 */
+				dom->accesslog = 1;
+			} else {
+				dom->accesslog = 0;
+			}
 		}
 	}
 }
@@ -555,13 +593,10 @@ kore_domain_closelogs(void)
 void
 kore_domain_keymgr_init(void)
 {
-#if !defined(KORE_NO_TLS)
 	keymgr_init();
 	kore_msg_register(KORE_MSG_KEYMGR_RESP, keymgr_msg_response);
-#endif
 }
 
-#if !defined(KORE_NO_TLS)
 static void
 keymgr_init(void)
 {
@@ -570,7 +605,7 @@ keymgr_init(void)
 	if ((meth = RSA_get_default_method()) == NULL)
 		fatal("failed to obtain RSA method");
 
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	RSA_meth_set_pub_enc(keymgr_rsa_meth, RSA_meth_get_pub_enc(meth));
 	RSA_meth_set_pub_dec(keymgr_rsa_meth, RSA_meth_get_pub_dec(meth));
 	RSA_meth_set_bn_mod_exp(keymgr_rsa_meth, RSA_meth_get_bn_mod_exp(meth));
@@ -585,7 +620,7 @@ static int
 keymgr_rsa_init(RSA *rsa)
 {
 	if (rsa != NULL) {
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 		RSA_set_flags(rsa, RSA_flags(rsa) |
 		    RSA_FLAG_EXT_PKEY | RSA_METHOD_FLAG_NO_CHECK);
 #else
@@ -612,18 +647,18 @@ keymgr_rsa_privenc(int flen, const unsigned char *from, unsigned char *to,
 
 	if ((dom = RSA_get_app_data(rsa)) == NULL)
 		fatal("RSA key has no domain attached");
-	if (strlen(dom->domain) >= KORE_DOMAINNAME_LEN - 1)
-		fatal("domain name too long");
 
 	memset(keymgr_buf, 0, sizeof(keymgr_buf));
 
 	req = (struct kore_keyreq *)keymgr_buf;
+
+	if (kore_strlcpy(req->domain, dom->domain, sizeof(req->domain)) >=
+	    sizeof(req->domain))
+		fatal("%s: domain truncated", __func__);
+
 	req->data_len = flen;
 	req->padding = padding;
-	req->domain_len = strlen(dom->domain);
-
 	memcpy(&req->data[0], from, req->data_len);
-	memcpy(req->domain, dom->domain, req->domain_len);
 
 	kore_msg_send(KORE_WORKER_KEYMGR, KORE_MSG_KEYMGR_REQ, keymgr_buf, len);
 	keymgr_await_data();
@@ -667,7 +702,7 @@ keymgr_ecdsa_sign(const unsigned char *dgst, int dgst_len,
 	if (len > sizeof(keymgr_buf))
 		fatal("keymgr_buf too small");
 
-#if !defined(LIBRESSL_VERSION_TEXT) && OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if defined(KORE_OPENSSL_NEWER_API)
 	if ((dom = EC_KEY_get_ex_data(eckey, 0)) == NULL)
 		fatal("EC_KEY has no domain");
 #else
@@ -676,13 +711,14 @@ keymgr_ecdsa_sign(const unsigned char *dgst, int dgst_len,
 #endif
 
 	memset(keymgr_buf, 0, sizeof(keymgr_buf));
-
 	req = (struct kore_keyreq *)keymgr_buf;
-	req->data_len = dgst_len;
-	req->domain_len = strlen(dom->domain);
 
+	if (kore_strlcpy(req->domain, dom->domain, sizeof(req->domain)) >=
+	    sizeof(req->domain))
+		fatal("%s: domain truncated", __func__);
+
+	req->data_len = dgst_len;
 	memcpy(&req->data[0], dgst, req->data_len);
-	memcpy(req->domain, dom->domain, req->domain_len);
 
 	kore_msg_send(KORE_WORKER_KEYMGR, KORE_MSG_KEYMGR_REQ, keymgr_buf, len);
 	keymgr_await_data();
@@ -845,31 +881,31 @@ domain_load_certificate_chain(SSL_CTX *ctx, const void *data, size_t len)
 	X509		*x, *ca;
 
 	ERR_clear_error();
-	in = domain_bio_mem(data, len);
+	in = BIO_new_mem_buf(data, len);
 
 	if ((x = PEM_read_bio_X509_AUX(in, NULL, NULL, NULL)) == NULL)
-		fatal("PEM_read_bio_X509_AUX: %s", ssl_errno_s);
+		return (NULL);
 
 	/* refcount for x509 will go up one. */
 	if (SSL_CTX_use_certificate(ctx, x) == 0)
-		fatal("SSL_CTX_use_certificate: %s", ssl_errno_s);
+		return (NULL);
 
-#if defined(LIBRESSL_VERSION_TEXT)
+#if defined(KORE_OPENSSL_NEWER_API)
+	SSL_CTX_clear_chain_certs(ctx);
+#else
 	sk_X509_pop_free(ctx->extra_certs, X509_free);
 	ctx->extra_certs = NULL;
-#else
-	SSL_CTX_clear_chain_certs(ctx);
 #endif
 
 	ERR_clear_error();
 	while ((ca = PEM_read_bio_X509(in, NULL, NULL, NULL)) != NULL) {
 		/* ca its reference count won't be increased. */
-#if defined(LIBRESSL_VERSION_TEXT)
-		if (SSL_CTX_add_extra_chain_cert(ctx, ca) == 0)
-			fatal("SSL_CTX_add_extra_chain_cert: %s", ssl_errno_s);
-#else
+#if defined(KORE_OPENSSL_NEWER_API)
 		if (SSL_CTX_add0_chain_cert(ctx, ca) == 0)
-			fatal("SSL_CTX_add0_chain_cert: %s", ssl_errno_s);
+			return (NULL);
+#else
+		if (SSL_CTX_add_extra_chain_cert(ctx, ca) == 0)
+			return (NULL);
 #endif
 	}
 
@@ -877,40 +913,9 @@ domain_load_certificate_chain(SSL_CTX *ctx, const void *data, size_t len)
 
 	if (ERR_GET_LIB(err) != ERR_LIB_PEM ||
 	    ERR_GET_REASON(err) != PEM_R_NO_START_LINE)
-		fatal("PEM_read_bio_X509: %s", ssl_errno_s);
+		return (NULL);
 
 	BIO_free(in);
 
 	return (x);
 }
-
-/*
- * XXX - Hack around the fact that LibreSSL its BIO_new_mem_buf() does not
- * take a const pointer for their first argument.
- *
- * Since we build with -Wcast-qual and -Werror I rather do this than having
- * a bunch of pragma preprocessor magic to remove the warnings for that code
- * if we're dealing with LibreSSL.
- *
- * They fixed this in their upcoming 2.8.0 release but that is not out yet
- * and I'd like this to run on older OpenBSD platforms as well.
- */
-static BIO *
-domain_bio_mem(const void *data, size_t len)
-{
-	BIO					*in;
-	union { void *p; const void *cp; }	deconst;
-
-	/* because OpenSSL likes taking ints as memory buffer lengths. */
-	if (len > INT_MAX)
-		fatal("domain_bio_mem: len(%zu) > INT_MAX", len);
-
-	deconst.cp = data;
-
-	if ((in = BIO_new_mem_buf(deconst.p, len)) == NULL)
-		fatal("BIO_new_mem_buf: %s", ssl_errno_s);
-
-	return (in);
-}
-
-#endif
