@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -91,7 +91,7 @@ static const char http_token[] = {
 	'`' , 'a' , 'b' , 'c' , 'd' , 'e' , 'f' , 'g' ,
 	'h' , 'i' , 'j' , 'k' , 'l' , 'm' , 'n' , 'o' ,
 	'p' , 'q' , 'r' , 's' , 't' , 'u' , 'v' , 'w' ,
-	'x' , 'y' , 'z' , 0x00, '|' , 0x00, '~' ,
+	'x' , 'y' , 'z' , 0x00, '|' , 0x00, '~' , 0x00
 };
 
 /*
@@ -115,14 +115,25 @@ static const char http_field_content[] = {
 	'`' , 'a' , 'b' , 'c' , 'd' , 'e' , 'f' , 'g' ,
 	'h' , 'i' , 'j' , 'k' , 'l' , 'm' , 'n' , 'o' ,
 	'p' , 'q' , 'r' , 's' , 't' , 'u' , 'v' , 'w' ,
-	'x' , 'y' , 'z' , '{' , '|' , '}' , '~' ,
+	'x' , 'y' , 'z' , '{' , '|' , '}' , '~' , 0x00
 };
 
+/*
+ * Fixed "pretty" HTTP error HTML page.
+ */
+static const char *pretty_error_fmt =
+	"<html>\n<head>\n\t<title>%d %s</title>"
+	"</head>\n<body>\n\t"
+	"<h1>%d %s</h1>\n"
+	"</body>\n</html>\n";
+
 static int	http_body_recv(struct netbuf *);
 static void	http_error_response(struct connection *, int);
 static void	http_write_response_cookie(struct http_cookie *);
 static void	http_argument_add(struct http_request *, char *, char *,
 		    int, int);
+static int	http_check_redirect(struct http_request *,
+		    struct kore_domain *);
 static void	http_response_normal(struct http_request *,
 		    struct connection *, int, const void *, size_t);
 static void	multipart_add_field(struct http_request *, struct kore_buf *,
@@ -153,6 +164,7 @@ static struct kore_pool			http_rlq_pool;
 
 struct kore_pool			http_header_pool;
 
+int		http_pretty_error = 0;
 u_int32_t	http_request_count = 0;
 u_int32_t	http_request_ms = HTTP_REQUEST_MS;
 u_int16_t	http_body_timeout = HTTP_BODY_TIMEOUT;
@@ -183,12 +195,14 @@ http_init(void)
 	header_buf = kore_buf_alloc(HTTP_HEADER_BUFSIZE);
 	ckhdr_buf = kore_buf_alloc(HTTP_COOKIE_BUFSIZE);
 
-	l = snprintf(http_version, sizeof(http_version),
-	    "server: kore (%s)\r\n", kore_version);
-	if (l == -1 || (size_t)l >= sizeof(http_version))
-		fatal("http_init(): http_version buffer too small");
+	if (!http_version_len) {
+		l = snprintf(http_version, sizeof(http_version),
+		    "server: kore (%s)\r\n", kore_version);
+		if (l == -1 || (size_t)l >= sizeof(http_version))
+			fatal("http_init(): http_version buffer too small");
 
-	http_version_len = l;
+		http_version_len = l;
+	}
 
 	prealloc = MIN((worker_max_connections / 10), 1000);
 	kore_pool_init(&http_request_pool, "http_request_pool",
@@ -442,6 +456,11 @@ http_request_free(struct http_request *req)
 		kore_python_coro_delete(req->py_coro);
 		req->py_coro = NULL;
 	}
+	if (req->py_validator != NULL) {
+		kore_python_coro_delete(req->py_validator);
+		req->py_validator = NULL;
+	}
+	Py_XDECREF(req->py_req);
 #endif
 #if defined(KORE_USE_PGSQL)
 	while (!LIST_EMPTY(&(req->pgsqls))) {
@@ -983,7 +1002,8 @@ http_argument_urldecode(char *arg)
 			return (KORE_RESULT_ERROR);
 		}
 
-		if (!isxdigit(*(p + 1)) || !isxdigit(*(p + 2))) {
+		if (!isxdigit((unsigned char)*(p + 1)) ||
+		    !isxdigit((unsigned char)*(p + 2))) {
 			*in++ = *p++;
 			continue;
 		}
@@ -998,7 +1018,7 @@ http_argument_urldecode(char *arg)
 		if (err != KORE_RESULT_OK)
 			return (err);
 
-		if (v <= 0x1f || v == 0x7f || (v >= 0x80 && v <= 0x9f))
+		if (v <= 0x1f || v == 0x7f)
 			return (KORE_RESULT_ERROR);
 
 		*in++ = (char)v;
@@ -1477,12 +1497,94 @@ http_runlock_release(struct http_runlock *lock, struct http_request *req)
 	}
 }
 
+int
+http_redirect_add(struct kore_domain *dom, const char *path, int status,
+    const char *target)
+{
+	struct http_redirect	*rdr;
+
+	rdr = kore_calloc(1, sizeof(*rdr));
+
+	if (regcomp(&(rdr->rctx), path, REG_EXTENDED)) {
+		kore_free(rdr);
+		return (KORE_RESULT_ERROR);
+	}
+
+	rdr->status = status;
+
+	if (target != NULL)
+		rdr->target = kore_strdup(target);
+	else
+		rdr->target = NULL;
+
+	TAILQ_INSERT_TAIL(&dom->redirects, rdr, list);
+
+	return (KORE_RESULT_OK);
+}
+
+static int
+http_check_redirect(struct http_request *req, struct kore_domain *dom)
+{
+	int			idx;
+	struct http_redirect	*rdr;
+	const char		*uri;
+	char			key[4];
+	struct kore_buf		location;
+
+	TAILQ_FOREACH(rdr, &dom->redirects, list) {
+		if (!regexec(&(rdr->rctx), req->path,
+		    HTTP_CAPTURE_GROUPS, req->cgroups, 0))
+			break;
+	}
+
+	if (rdr == NULL)
+		return (KORE_RESULT_ERROR);
+
+	uri = NULL;
+	kore_buf_init(&location, 128);
+
+	if (rdr->target) {
+		kore_buf_appendf(&location, "%s", rdr->target);
+
+		if (req->query_string != NULL) {
+			kore_buf_replace_string(&location, "$qs",
+			    req->query_string, strlen(req->query_string));
+		}
+
+		/* Starts at 1 to skip the full path. */
+		for (idx = 1; idx < HTTP_CAPTURE_GROUPS - 1; idx++) {
+			if (req->cgroups[idx].rm_so == -1 ||
+			    req->cgroups[idx].rm_eo == -1)
+				break;
+
+			(void)snprintf(key, sizeof(key), "$%d", idx);
+
+			kore_buf_replace_string(&location, key,
+			    req->path + req->cgroups[idx].rm_so,
+			    req->cgroups[idx].rm_eo - req->cgroups[idx].rm_so);
+		}
+
+		uri = kore_buf_stringify(&location, NULL);
+	}
+
+	if (uri)
+		http_response_header(req, "location", uri);
+
+	http_response(req, rdr->status, NULL, 0);
+	kore_buf_cleanup(&location);
+
+	if (dom->accesslog)
+		kore_accesslog(req);
+
+	return (KORE_RESULT_OK);
+}
+
 static struct http_request *
 http_request_new(struct connection *c, const char *host,
     const char *method, char *path, const char *version)
 {
+	struct kore_domain		*dom;
 	struct http_request		*req;
-	struct kore_module_handle	*hdlr;
 	char				*p, *hp;
 	int				m, flags;
 	size_t				hostlen, pathlen, qsoff;
@@ -1517,7 +1619,6 @@ http_request_new(struct connection *c, const char *host,
 	}
 
 	if ((p = strchr(path, '?')) != NULL) {
-		*p = '\0';
 		qsoff = p - path;
 	} else {
 		qsoff = 0;
@@ -1545,17 +1646,26 @@ http_request_new(struct connection *c, const char *host,
 		break;
 	}
 
-	if ((hdlr = kore_module_handler_find(host, path)) == NULL) {
-		http_error_response(c, 404);
+	if (c->owner->server->tls && c->tls_sni != NULL) {
+		if (strcasecmp(c->tls_sni, host)) {
+			http_error_response(c, HTTP_STATUS_MISDIRECTED_REQUEST);
+			return (NULL);
+		}
+	}
+
+	if ((dom = kore_domain_lookup(c->owner->server, host)) == NULL) {
+		http_error_response(c, HTTP_STATUS_NOT_FOUND);
+		return (NULL);
+	}
+
+	if (dom->cafile != NULL && c->cert == NULL) {
+		http_error_response(c, HTTP_STATUS_FORBIDDEN);
 		return (NULL);
 	}
 
 	if (hp != NULL)
 		*hp = ':';
 
-	if (p != NULL)
-		*p = '?';
-
 	if (!strcasecmp(method, "get")) {
 		m = HTTP_METHOD_GET;
 		flags |= HTTP_REQUEST_COMPLETE;
@@ -1590,19 +1700,14 @@ http_request_new(struct connection *c, const char *host,
 		}
 	}
 
-	if (!(hdlr->methods & m)) {
-		http_error_response(c, HTTP_STATUS_METHOD_NOT_ALLOWED);
-		return (NULL);
-	}
-
 	req = kore_pool_get(&http_request_pool);
+
 	req->end = 0;
 	req->total = 0;
 	req->start = 0;
 	req->owner = c;
 	req->status = 0;
 	req->method = m;
-	req->hdlr = hdlr;
 	req->agent = NULL;
 	req->onfree = NULL;
 	req->referer = NULL;
@@ -1621,7 +1726,10 @@ http_request_new(struct connection *c, const char *host,
 	req->path = path;
 
 #if defined(KORE_USE_PYTHON)
+	req->py_req = NULL;
 	req->py_coro = NULL;
+	req->py_rqnext = NULL;
+	req->py_validator = NULL;
 #endif
 
 	if (qsoff > 0) {
@@ -1631,6 +1739,9 @@ http_request_new(struct connection *c, const char *host,
 		req->query_string = NULL;
 	}
 
+	/* Checked further down below if we need to 404. */
+	req->hdlr = kore_module_handler_find(req, dom);
+
 	TAILQ_INIT(&(req->resp_headers));
 	TAILQ_INIT(&(req->req_headers));
 	TAILQ_INIT(&(req->resp_cookies));
@@ -1650,6 +1761,23 @@ http_request_new(struct connection *c, const char *host,
 	TAILQ_INSERT_HEAD(&http_requests, req, list);
 	TAILQ_INSERT_TAIL(&(c->http_requests), req, olist);
 
+	if (http_check_redirect(req, dom)) {
+		http_request_free(req);
+		return (NULL);
+	}
+
+	if (req->hdlr == NULL) {
+		http_request_free(req);
+		http_error_response(c, HTTP_STATUS_NOT_FOUND);
+		return (NULL);
+	}
+
+	if (!(req->hdlr->methods & m)) {
+		http_request_free(req);
+		http_error_response(c, HTTP_STATUS_METHOD_NOT_ALLOWED);
+		return (NULL);
+	}
+
 	return (req);
 }
 
@@ -1850,8 +1978,10 @@ http_argument_add(struct http_request *req, char *name, char *value, int qs,
 	struct http_arg			*q;
 	struct kore_handler_params	*p;
 
-	if (decode)
-		http_argument_urldecode(name);
+	if (decode) {
+		if (!http_argument_urldecode(name))
+			return;
+	}
 
 	TAILQ_FOREACH(p, &(req->hdlr->params), list) {
 		if (qs == 1 && !(p->flags & KORE_PARAMS_QUERY_STRING))
@@ -1865,8 +1995,10 @@ http_argument_add(struct http_request *req, char *name, char *value, int qs,
 		if (strcmp(p->name, name))
 			continue;
 
-		if (decode)
-			http_argument_urldecode(value);
+		if (decode) {
+			if (!http_argument_urldecode(value))
+				return;
+		}
 
 		if (!kore_validator_check(req, p->validator, value))
 			break;
@@ -1951,12 +2083,17 @@ static void
 http_response_normal(struct http_request *req, struct connection *c,
     int status, const void *d, size_t len)
 {
+	struct kore_buf		buf;
 	struct http_cookie	*ck;
 	struct http_header	*hdr;
-	const char		*conn;
 	char			version;
-	int			connection_close;
+	const char		*conn, *text;
+	int			connection_close, send_body;
 
+	send_body = 1;
+	text = http_status_text(status);
+
+	kore_buf_init(&buf, 1024);
 	kore_buf_reset(header_buf);
 
 	if (req != NULL) {
@@ -1969,11 +2106,11 @@ http_response_normal(struct http_request *req, struct connection *c,
 	}
 
 	kore_buf_appendf(header_buf, "HTTP/1.%c %d %s\r\n",
-	    version, status, http_status_text(status));
+	    version, status, text);
 	kore_buf_append(header_buf, http_version, http_version_len);
 
 	if ((c->flags & CONN_CLOSE_EMPTY) ||
-	    (req->flags & HTTP_VERSION_1_0)) {
+	    (req != NULL && (req->flags & HTTP_VERSION_1_0))) {
 		connection_close = 1;
 	} else {
 		connection_close = 0;
@@ -2008,6 +2145,14 @@ http_response_normal(struct http_request *req, struct connection *c,
 		    http_hsts_enable);
 	}
 
+	if (http_pretty_error && d == NULL && status >= 400) {
+		kore_buf_appendf(&buf, pretty_error_fmt,
+		    status, text, status, text);
+
+		d = buf.data;
+		len = buf.offset;
+	}
+
 	if (req != NULL) {
 		TAILQ_FOREACH(ck, &(req->resp_cookies), list)
 			http_write_response_cookie(ck);
@@ -2032,7 +2177,10 @@ http_response_normal(struct http_request *req, struct connection *c,
 	kore_buf_append(header_buf, "\r\n", 2);
 	net_send_queue(c, header_buf->data, header_buf->offset);
 
-	if (d != NULL && req != NULL && req->method != HTTP_METHOD_HEAD)
+	if (req != NULL && req->method == HTTP_METHOD_HEAD)
+		send_body = 0;
+
+	if (d != NULL && send_body)
 		net_send_queue(c, d, len);
 
 	if (!(c->flags & CONN_CLOSE_EMPTY) && !(c->flags & CONN_IS_BUSY))
@@ -2040,6 +2188,8 @@ http_response_normal(struct http_request *req, struct connection *c,
 
 	if (req != NULL)
 		req->content_length = len;
+
+	kore_buf_cleanup(&buf);
 }
 
 static void
@@ -2191,6 +2341,9 @@ http_status_text(int status)
 	case HTTP_STATUS_EXPECTATION_FAILED:
 		r = "Expectation Failed";
 		break;
+	case HTTP_STATUS_MISDIRECTED_REQUEST:
+		r = "Misdirected Request";
+		break;
 	case HTTP_STATUS_INTERNAL_ERROR:
 		r = "Internal Server Error";
 		break;
@@ -2252,6 +2405,33 @@ http_method_text(int method)
 	return (r);
 }
 
+int
+http_method_value(const char *method)
+{
+	if (!strcasecmp(method, "GET"))
+		return (HTTP_METHOD_GET);
+
+	if (!strcasecmp(method, "POST"))
+		return (HTTP_METHOD_POST);
+
+	if (!strcasecmp(method, "PUT"))
+		return (HTTP_METHOD_PUT);
+
+	if (!strcasecmp(method, "DELETE"))
+		return (HTTP_METHOD_DELETE);
+
+	if (!strcasecmp(method, "HEAD"))
+		return (HTTP_METHOD_HEAD);
+
+	if (!strcasecmp(method, "OPTIONS"))
+		return (HTTP_METHOD_OPTIONS);
+
+	if (!strcasecmp(method, "PATCH"))
+		return (HTTP_METHOD_PATCH);
+
+	return (0);
+}
+
 int
 http_media_register(const char *ext, const char *type)
 {
