@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -37,12 +37,20 @@
 #include "curl.h"
 #endif
 
+#if defined(KORE_USE_PGSQL)
+#include "pgsql.h"
+#endif
+
 #if defined(KORE_USE_PYTHON)
 #include "python_api.h"
 #endif
 
+#if defined(KORE_USE_ACME)
+#include "acme.h"
+#endif
+
 volatile sig_atomic_t	sig_recv;
-struct listener_head	listeners;
+struct kore_server_list	kore_servers;
 u_int8_t		nlisteners;
 int			kore_argc = 0;
 pid_t			kore_pid = -1;
@@ -71,31 +79,47 @@ static void	kore_write_kore_pid(void);
 static void	kore_proctitle_setup(void);
 static void	kore_server_sslstart(void);
 static void	kore_server_start(int, char *[]);
+static void	kore_call_parent_configure(int, char **);
+
+#if !defined(KORE_SINGLE_BINARY) && defined(KORE_USE_PYTHON)
+static const char	*parent_config_hook = KORE_PYTHON_CONFIG_HOOK;
+static const char	*parent_teardown_hook = KORE_PYTHON_TEARDOWN_HOOK;
+#else
+static const char	*parent_config_hook = KORE_CONFIG_HOOK;
+static const char	*parent_teardown_hook = KORE_TEARDOWN_HOOK;
+#if defined(KORE_SINGLE_BINARY)
+static const char	*parent_daemonized_hook = KORE_DAEMONIZED_HOOK;
+#endif
+#endif
 
 static void
 usage(void)
 {
-	fprintf(stderr, "Usage: %s [options]\n", __progname);
+#if !defined(KORE_SINGLE_BINARY) && defined(KORE_USE_PYTHON)
+	printf("Usage: %s [options] [app | app.py]\n", __progname);
+#else
+	printf("Usage: %s [options]\n", __progname);
+#endif
 
-	fprintf(stderr, "\n");
-	fprintf(stderr, "Available options:\n");
+	printf("\n");
+	printf("Available options:\n");
 #if !defined(KORE_SINGLE_BINARY)
-	fprintf(stderr, "\t-c\tconfiguration to use\n");
+	printf("\t-c\tconfiguration to use\n");
 #endif
 #if defined(KORE_DEBUG)
-	fprintf(stderr, "\t-d\trun with debug on\n");
+	printf("\t-d\trun with debug on\n");
 #endif
-	fprintf(stderr, "\t-f\tstart in foreground\n");
-	fprintf(stderr, "\t-h\tthis help text\n");
-	fprintf(stderr, "\t-n\tdo not chroot\n");
-	fprintf(stderr, "\t-q\tonly log errors\n");
-	fprintf(stderr, "\t-r\tdo not drop privileges\n");
-	fprintf(stderr, "\t-v\tdisplay %s build information\n", __progname);
+	printf("\t-f\tstart in foreground\n");
+	printf("\t-h\tthis help text\n");
+	printf("\t-n\tdo not chroot\n");
+	printf("\t-q\tonly log errors\n");
+	printf("\t-r\tdo not drop privileges\n");
+	printf("\t-v\tdisplay %s build information\n", __progname);
 
 #if !defined(KORE_SINGLE_BINARY)
-	fprintf(stderr, "\nFind more information on https://kore.io\n");
+	printf("\nFind more information on https://kore.io\n");
 #else
-	fprintf(stderr, "\nBuilt using https://kore.io\n");
+	printf("\nBuilt using https://kore.io\n");
 #endif
 
 	exit(1);
@@ -105,14 +129,11 @@ static void
 version(void)
 {
 	printf("%s ", kore_version);
-#if defined(KORE_NO_TLS)
-	printf("no-tls ");
-#endif
 #if defined(KORE_NO_HTTP)
 	printf("no-http ");
 #endif
 #if defined(KORE_USE_CURL)
-	printf("curl ");
+	printf("curl-%s ", LIBCURL_VERSION);
 #endif
 #if defined(KORE_USE_PGSQL)
 	printf("pgsql ");
@@ -127,7 +148,10 @@ version(void)
 	printf("single ");
 #endif
 #if defined(KORE_USE_PYTHON)
-	printf("python ");
+	printf("python-%s ", PY_VERSION);
+#endif
+#if defined(KORE_USE_ACME)
+	printf("acme ");
 #endif
 	printf("\n");
 	exit(0);
@@ -138,6 +162,9 @@ main(int argc, char *argv[])
 {
 	struct kore_runtime_call	*rcall;
 	int				ch, flags;
+#if !defined(KORE_SINGLE_BINARY) && defined(KORE_USE_PYTHON)
+	struct stat			st;
+#endif
 
 	flags = 0;
 
@@ -153,7 +180,9 @@ main(int argc, char *argv[])
 		switch (ch) {
 #if !defined(KORE_SINGLE_BINARY)
 		case 'c':
-			config_file = optarg;
+			free(config_file);
+			if ((config_file = strdup(optarg)) == NULL)
+				fatal("strdup");
 			break;
 #endif
 #if defined(KORE_DEBUG)
@@ -191,30 +220,57 @@ main(int argc, char *argv[])
 	argc -= optind;
 	argv += optind;
 
-#if !defined(KORE_SINGLE_BINARY)
+#if !defined(KORE_SINGLE_BINARY) && defined(KORE_USE_PYTHON)
+	if (argc > 0) {
+		kore_pymodule = argv[0];
+		argc--;
+		argv++;
+	} else {
+		kore_pymodule = NULL;
+	}
+
+	if (kore_pymodule) {
+		if (lstat(kore_pymodule, &st) == -1) {
+			fatal("failed to stat '%s': %s",
+			    kore_pymodule, errno_s);
+		}
+
+		if (!S_ISDIR(st.st_mode) && !S_ISREG(st.st_mode))
+			fatal("%s: not a directory or file", kore_pymodule);
+	}
+
+#elif !defined(KORE_SINGLE_BINARY)
 	if (argc > 0)
 		fatal("did you mean to run `kodev' instead?");
 #endif
 
 	kore_pid = getpid();
 	nlisteners = 0;
-	LIST_INIT(&listeners);
+	LIST_INIT(&kore_servers);
 
+	kore_platform_init();
 	kore_log_init();
+	kore_msg_init();
 #if !defined(KORE_NO_HTTP)
 	http_parent_init();
 #if defined(KORE_USE_CURL)
 	kore_curl_sysinit();
+#endif
+#if defined(KORE_USE_PGSQL)
+	kore_pgsql_sys_init();
 #endif
 	kore_auth_init();
 	kore_validator_init();
 	kore_filemap_init();
+#endif
+#if defined(KORE_USE_ACME)
+	kore_acme_init();
 #endif
 	kore_domain_init();
 	kore_module_init();
 	kore_server_sslstart();
 
-#if !defined(KORE_SINGLE_BINARY)
+#if !defined(KORE_SINGLE_BINARY) && !defined(KORE_USE_PYTHON)
 	if (config_file == NULL)
 		usage();
 #endif
@@ -222,20 +278,34 @@ main(int argc, char *argv[])
 
 #if defined(KORE_USE_PYTHON)
 	kore_python_init();
+#if !defined(KORE_SINGLE_BINARY)
+	if (kore_pymodule) {
+		kore_module_load(kore_pymodule, NULL, KORE_MODULE_PYTHON);
+		if (S_ISDIR(st.st_mode) && chdir(kore_pymodule) == -1)
+			fatal("chdir(%s): %s", kore_pymodule, errno_s);
+	} else {
+		/* swap back to non-python hooks. */
+		parent_config_hook = KORE_CONFIG_HOOK;
+		parent_teardown_hook = KORE_TEARDOWN_HOOK;
+	}
+#endif
+#endif
+
+#if defined(KORE_SINGLE_BINARY)
+	kore_call_parent_configure(argc, argv);
+#endif
+
+#if defined(KORE_USE_PYTHON) && !defined(KORE_SINGLE_BINARY)
+	if (kore_pymodule)
+		kore_call_parent_configure(argc, argv);
 #endif
 
 	kore_parse_config();
 
-#if defined(KORE_SINGLE_BINARY)
-	rcall = kore_runtime_getcall("kore_parent_configure");
-	if (rcall != NULL) {
-		kore_runtime_configure(rcall, argc, argv);
-		kore_free(rcall);
-	}
+#if !defined(KORE_SINGLE_BINARY)
+	free(config_file);
 #endif
 
-	kore_platform_init();
-
 #if !defined(KORE_NO_HTTP)
 	if (http_body_disk_offload > 0) {
 		if (mkdir(http_body_disk_path, 0700) == -1 && errno != EEXIST) {
@@ -254,7 +324,7 @@ main(int argc, char *argv[])
 
 	kore_worker_shutdown();
 
-	rcall = kore_runtime_getcall("kore_parent_teardown");
+	rcall = kore_runtime_getcall(parent_teardown_hook);
 	if (rcall != NULL) {
 		kore_runtime_execute(rcall);
 		kore_free(rcall);
@@ -263,7 +333,7 @@ main(int argc, char *argv[])
 	if (unlink(kore_pidfile) == -1 && errno != ENOENT)
 		kore_log(LOG_NOTICE, "failed to remove pidfile (%s)", errno_s);
 
-	kore_listener_cleanup();
+	kore_server_cleanup();
 
 	if (!kore_quiet)
 		kore_log(LOG_NOTICE, "goodbye");
@@ -277,17 +347,24 @@ main(int argc, char *argv[])
 	return (0);
 }
 
-#if !defined(KORE_NO_TLS)
 int
 kore_tls_sni_cb(SSL *ssl, int *ad, void *arg)
 {
+	struct connection	*c;
 	struct kore_domain	*dom;
 	const char		*sname;
 
+	if ((c = SSL_get_ex_data(ssl, 0)) == NULL)
+		fatal("no connection data in %s", __func__);
+
 	sname = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
 	kore_debug("kore_tls_sni_cb(): received host %s", sname);
 
-	if (sname != NULL && (dom = kore_domain_lookup(sname)) != NULL) {
+	if (sname != NULL)
+		c->tls_sni = kore_strdup(sname);
+
+	if (sname != NULL &&
+	    (dom = kore_domain_lookup(c->owner->server, sname)) != NULL) {
 		if (dom->ssl_ctx == NULL) {
 			kore_log(LOG_NOTICE,
 			    "TLS configuration for %s not complete",
@@ -301,6 +378,7 @@ kore_tls_sni_cb(SSL *ssl, int *ad, void *arg)
 		if (dom->cafile != NULL) {
 			SSL_set_verify(ssl, SSL_VERIFY_PEER |
 			    SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
+			c->flags |= CONN_LOG_TLS_FAILURE;
 		} else {
 			SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);
 		}
@@ -326,16 +404,16 @@ kore_tls_info_callback(const SSL *ssl, int flags, int ret)
 			c->tls_reneg++;
 	}
 }
-#endif
 
 int
-kore_server_bind(const char *ip, const char *port, const char *ccb)
+kore_server_bind(struct kore_server *srv, const char *ip, const char *port,
+    const char *ccb)
 {
 	int			r;
 	struct listener		*l;
 	struct addrinfo		hints, *results;
 
-	kore_debug("kore_server_bind(%s, %s)", ip, port);
+	kore_debug("kore_server_bind(%s, %s, %s)", srv->name, ip, port);
 
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_family = AF_UNSPEC;
@@ -347,7 +425,11 @@ kore_server_bind(const char *ip, const char *port, const char *ccb)
 	if (r != 0)
 		fatal("getaddrinfo(%s): %s", ip, gai_strerror(r));
 
-	if ((l = kore_listener_alloc(results->ai_family, ccb)) == NULL) {
+	l = kore_listener_create(srv);
+	l->host = kore_strdup(ip);
+	l->port = kore_strdup(port);
+
+	if (!kore_listener_init(l, results->ai_family, ccb)) {
 		freeaddrinfo(results);
 		return (KORE_RESULT_ERROR);
 	}
@@ -367,19 +449,12 @@ kore_server_bind(const char *ip, const char *port, const char *ccb)
 		return (KORE_RESULT_ERROR);
 	}
 
-	if (foreground && !kore_quiet) {
-#if !defined(KORE_NO_TLS)
-		kore_log(LOG_NOTICE, "running on https://%s:%s", ip, port);
-#else
-		kore_log(LOG_NOTICE, "running on http://%s:%s", ip, port);
-#endif
-	}
-
 	return (KORE_RESULT_OK);
 }
 
 int
-kore_server_bind_unix(const char *path, const char *ccb)
+kore_server_bind_unix(struct kore_server *srv, const char *path,
+    const char *ccb)
 {
 	struct listener		*l;
 	int			len;
@@ -403,7 +478,10 @@ kore_server_bind_unix(const char *path, const char *ccb)
 	socklen = sizeof(sun);
 #endif
 
-	if ((l = kore_listener_alloc(AF_UNIX, ccb)) == NULL)
+	l = kore_listener_create(srv);
+	l->host = kore_strdup(path);
+
+	if (!kore_listener_init(l, AF_UNIX, ccb))
 		return (KORE_RESULT_ERROR);
 
 	if (bind(l->fd, (struct sockaddr *)&sun, socklen) == -1) {
@@ -418,17 +496,86 @@ kore_server_bind_unix(const char *path, const char *ccb)
 		return (KORE_RESULT_ERROR);
 	}
 
-	if (foreground && !kore_quiet)
-		kore_log(LOG_NOTICE, "running on %s", path);
-
 	return (KORE_RESULT_OK);
 }
 
+struct kore_server *
+kore_server_create(const char *name)
+{
+	struct kore_server	*srv;
+
+	srv = kore_calloc(1, sizeof(struct kore_server));
+	srv->name = kore_strdup(name);
+	srv->tls = 1;
+
+	TAILQ_INIT(&srv->domains);
+	LIST_INIT(&srv->listeners);
+
+	LIST_INSERT_HEAD(&kore_servers, srv, list);
+
+	return (srv);
+}
+
+void
+kore_server_finalize(struct kore_server *srv)
+{
+	struct listener		*l;
+	const char		*proto;
+
+	if (kore_quiet)
+		return;
+
+	LIST_FOREACH(l, &srv->listeners, list) {
+		if (srv->tls)
+			proto = "https";
+		else
+			proto = "http";
+
+		if (l->family == AF_UNIX) {
+			kore_log(LOG_NOTICE, "%s serving %s on %s",
+			    srv->name, proto, l->host);
+		} else {
+			kore_log(LOG_NOTICE, "%s serving %s on %s:%s",
+			    srv->name, proto, l->host, l->port);
+		}
+	}
+}
+
 struct listener *
-kore_listener_alloc(int family, const char *ccb)
+kore_listener_create(struct kore_server *server)
 {
 	struct listener		*l;
 
+	l = kore_calloc(1, sizeof(struct listener));
+
+	nlisteners++;
+	LIST_INSERT_HEAD(&server->listeners, l, list);
+
+	l->server = server;
+
+	l->fd = -1;
+	l->evt.type = KORE_TYPE_LISTENER;
+	l->evt.handle = kore_listener_accept;
+
+	return (l);
+}
+
+struct kore_server *
+kore_server_lookup(const char *name)
+{
+	struct kore_server	*srv;
+
+	LIST_FOREACH(srv, &kore_servers, list) {
+		if (!strcmp(srv->name, name))
+			return (srv);
+	}
+
+	return (NULL);
+}
+
+int
+kore_listener_init(struct listener *l, int family, const char *ccb)
+{
 	switch (family) {
 	case AF_INET:
 	case AF_INET6:
@@ -438,51 +585,60 @@ kore_listener_alloc(int family, const char *ccb)
 		fatal("unknown address family %d", family);
 	}
 
-	l = kore_calloc(1, sizeof(struct listener));
-
-	nlisteners++;
-	LIST_INSERT_HEAD(&listeners, l, list);
-
-	l->fd = -1;
 	l->family = family;
 
-	l->evt.type = KORE_TYPE_LISTENER;
-	l->evt.handle = kore_listener_accept;
-
 	if ((l->fd = socket(family, SOCK_STREAM, 0)) == -1) {
 		kore_listener_free(l);
 		kore_log(LOG_ERR, "socket(): %s", errno_s);
-		return (NULL);
+		return (KORE_RESULT_ERROR);
 	}
 
 	if (fcntl(l->fd, F_SETFD, FD_CLOEXEC) == -1) {
 		kore_listener_free(l);
 		kore_log(LOG_ERR, "fcntl(): %s", errno_s);
-		return (NULL);
+		return (KORE_RESULT_ERROR);
 	}
 
 	if (!kore_connection_nonblock(l->fd, family != AF_UNIX)) {
 		kore_listener_free(l);
 		kore_log(LOG_ERR, "kore_connection_nonblock(): %s", errno_s);
-		return (NULL);
+		return (KORE_RESULT_ERROR);
 	}
 
 	if (!kore_sockopt(l->fd, SOL_SOCKET, SO_REUSEADDR)) {
 		kore_listener_free(l);
-		return (NULL);
+		return (KORE_RESULT_ERROR);
 	}
 
 	if (ccb != NULL) {
 		if ((l->connect = kore_runtime_getcall(ccb)) == NULL) {
 			kore_log(LOG_ERR, "no such callback: '%s'", ccb);
 			kore_listener_free(l);
-			return (NULL);
+			return (KORE_RESULT_ERROR);
 		}
 	} else {
 		l->connect = NULL;
 	}
 
-	return (l);
+	return (KORE_RESULT_OK);
+}
+
+void
+kore_server_free(struct kore_server *srv)
+{
+	struct listener		*l;
+	struct kore_domain	*dom;
+
+	LIST_REMOVE(srv, list);
+
+	while ((dom = TAILQ_FIRST(&srv->domains)) != NULL)
+		kore_domain_free(dom);
+
+	while ((l = LIST_FIRST(&srv->listeners)) != NULL)
+		kore_listener_free(l);
+
+	kore_free(srv->name);
+	kore_free(srv);
 }
 
 void
@@ -493,6 +649,9 @@ kore_listener_free(struct listener *l)
 	if (l->fd != -1)
 		close(l->fd);
 
+	kore_free(l->host);
+	kore_free(l->port);
+
 	kore_free(l);
 }
 
@@ -577,16 +736,26 @@ kore_signal_setup(void)
 }
 
 void
-kore_listener_cleanup(void)
+kore_server_closeall(void)
 {
 	struct listener		*l;
+	struct kore_server	*srv;
 
-	while (!LIST_EMPTY(&listeners)) {
-		l = LIST_FIRST(&listeners);
-		kore_listener_free(l);
+	LIST_FOREACH(srv, &kore_servers, list) {
+		LIST_FOREACH(l, &srv->listeners, list)
+			l->fd = -1;
 	}
 }
 
+void
+kore_server_cleanup(void)
+{
+	struct kore_server	*srv;
+
+	while ((srv = LIST_FIRST(&kore_servers)) != NULL)
+		kore_server_free(srv);
+}
+
 void
 kore_signal(int sig)
 {
@@ -628,7 +797,8 @@ kore_proctitle_setup(void)
 	proctitle_maxlen = 0;
 
 	for (i = 0; environ[i] != NULL; i++) {
-		p = kore_strdup(environ[i]);
+		if ((p = strdup(environ[i])) == NULL)
+			fatal("strdup");
 		proctitle_maxlen += strlen(environ[i]) + 1;
 		environ[i] = p;
 	}
@@ -640,27 +810,26 @@ kore_proctitle_setup(void)
 static void
 kore_server_sslstart(void)
 {
-#if !defined(KORE_NO_TLS)
-	kore_debug("kore_server_sslstart()");
-
 	SSL_library_init();
 	SSL_load_error_strings();
-#endif
 }
 
 static void
 kore_server_start(int argc, char *argv[])
 {
 	u_int32_t			tmp;
-	struct kore_runtime_call	*rcall;
+	struct kore_server		*srv;
 	u_int64_t			netwait;
 	int				quit, last_sig;
+#if defined(KORE_SINGLE_BINARY)
+	struct kore_runtime_call	*rcall;
+#endif
 
 	if (foreground == 0) {
 		if (daemon(1, 0) == -1)
 			fatal("cannot daemon(): %s", errno_s);
 #if defined(KORE_SINGLE_BINARY)
-		rcall = kore_runtime_getcall("kore_parent_daemonized");
+		rcall = kore_runtime_getcall(parent_daemonized_hook);
 		if (rcall != NULL) {
 			kore_runtime_execute(rcall);
 			kore_free(rcall);
@@ -673,6 +842,9 @@ kore_server_start(int argc, char *argv[])
 
 	if (!kore_quiet) {
 		kore_log(LOG_NOTICE, "%s is starting up", __progname);
+#if defined(__linux__)
+		kore_log(LOG_NOTICE, "seccomp sandbox enabled");
+#endif
 #if defined(KORE_USE_PGSQL)
 		kore_log(LOG_NOTICE, "pgsql built-in enabled");
 #endif
@@ -687,16 +859,24 @@ kore_server_start(int argc, char *argv[])
 #endif
 	}
 
-#if !defined(KORE_SINGLE_BINARY)
-	rcall = kore_runtime_getcall("kore_parent_configure");
-	if (rcall != NULL) {
-		kore_runtime_configure(rcall, argc, argv);
-		kore_free(rcall);
-	}
+#if !defined(KORE_SINGLE_BINARY) && !defined(KORE_USE_PYTHON)
+	kore_call_parent_configure(argc, argv);
 #endif
 
+#if defined(KORE_USE_PYTHON) && !defined(KORE_SINGLE_BINARY)
+	if (kore_pymodule == NULL)
+		kore_call_parent_configure(argc, argv);
+#endif
+
+	/* Check if keymgr will be active. */
+	LIST_FOREACH(srv, &kore_servers, list) {
+		if (srv->tls) {
+			keymgr_active = 1;
+			break;
+		}
+	}
+
 	kore_platform_proctitle("[parent]");
-	kore_msg_init();
 	kore_worker_init();
 
 	/* Set worker_max_connections for kore_connection_init(). */
@@ -716,6 +896,10 @@ kore_server_start(int argc, char *argv[])
 	kore_timer_add(kore_accesslog_run, 100, NULL, 0);
 #endif
 
+#if defined(KORE_USE_PYTHON)
+	kore_msg_unregister(KORE_PYTHON_SEND_OBJ);
+#endif
+
 	while (quit != 1) {
 		if (sig_recv != 0) {
 			last_sig = sig_recv;
@@ -776,3 +960,15 @@ kore_write_kore_pid(void)
 		fclose(fp);
 	}
 }
+
+static void
+kore_call_parent_configure(int argc, char **argv)
+{
+	struct kore_runtime_call	*rcall;
+
+	rcall = kore_runtime_getcall(parent_config_hook);
+	if (rcall != NULL) {
+		kore_runtime_configure(rcall, argc, argv);
+		kore_free(rcall);
+	}
+}
