@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2014-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -28,6 +28,22 @@
 
 #include "pgsql.h"
 
+#if defined(__linux__)
+#include "seccomp.h"
+
+static struct sock_filter filter_pgsql[] = {
+	/* Allow us to create sockets and call connect. */
+	KORE_SYSCALL_ALLOW(connect),
+	KORE_SYSCALL_ALLOW_ARG(socket, 0, AF_INET),
+	KORE_SYSCALL_ALLOW_ARG(socket, 0, AF_INET6),
+	KORE_SYSCALL_ALLOW_ARG(socket, 0, AF_UNIX),
+
+	/* Requires these calls. */
+	KORE_SYSCALL_ALLOW(getsockopt),
+	KORE_SYSCALL_ALLOW(getsockname),
+};
+#endif
+
 struct pgsql_wait {
 	struct kore_pgsql	*pgsql;
 	TAILQ_ENTRY(pgsql_wait)	list;
@@ -79,6 +95,11 @@ kore_pgsql_sys_init(void)
 	    sizeof(struct pgsql_job), 100);
 	kore_pool_init(&pgsql_wait_pool, "pgsql_wait_pool",
 	    sizeof(struct pgsql_wait), pgsql_queue_limit);
+
+#if defined(__linux__)
+	kore_seccomp_filter("pgsql", filter_pgsql,
+	    KORE_FILTER_LEN(filter_pgsql));
+#endif
 }
 
 void
@@ -323,12 +344,7 @@ kore_pgsql_handle(void *c, int err)
 
 	pgsql = conn->job->pgsql;
 
-	if (!PQconsumeInput(conn->db)) {
-		pgsql->state = KORE_PGSQL_STATE_ERROR;
-		pgsql->error = kore_strdup(PQerrorMessage(conn->db));
-	} else {
-		pgsql_read_result(pgsql);
-	}
+	pgsql_read_result(pgsql);
 
 	if (pgsql->state == KORE_PGSQL_STATE_WAIT) {
 #if !defined(KORE_NO_HTTP)
@@ -712,14 +728,33 @@ pgsql_conn_cleanup(struct pgsql_conn *conn)
 static void
 pgsql_read_result(struct kore_pgsql *pgsql)
 {
-	PGnotify	*notify;
+	struct pgsql_conn	*conn;
+	PGnotify		*notify;
+	int			saved_errno;
 
-	if (PQisBusy(pgsql->conn->db)) {
-		pgsql->state = KORE_PGSQL_STATE_WAIT;
-		return;
+	conn = pgsql->conn;
+
+	for (;;) {
+		if (!PQconsumeInput(conn->db)) {
+			pgsql->state = KORE_PGSQL_STATE_ERROR;
+			pgsql->error = kore_strdup(PQerrorMessage(conn->db));
+			return;
+		}
+
+		saved_errno = errno;
+
+		if (PQisBusy(conn->db)) {
+			if (saved_errno != EAGAIN && saved_errno != EWOULDBLOCK)
+				continue;
+			pgsql->state = KORE_PGSQL_STATE_WAIT;
+			conn->evt.flags &= ~KORE_EVENT_READ;
+			return;
+		}
+
+		break;
 	}
 
-	while ((notify = PQnotifies(pgsql->conn->db)) != NULL) {
+	while ((notify = PQnotifies(conn->db)) != NULL) {
 		pgsql->state = KORE_PGSQL_STATE_NOTIFY;
 		pgsql->notify.extra = notify->extra;
 		pgsql->notify.channel = notify->relname;
@@ -730,7 +765,7 @@ pgsql_read_result(struct kore_pgsql *pgsql)
 		PQfreemem(notify);
 	}
 
-	pgsql->result = PQgetResult(pgsql->conn->db);
+	pgsql->result = PQgetResult(conn->db);
 	if (pgsql->result == NULL) {
 		pgsql->state = KORE_PGSQL_STATE_DONE;
 		return;
