@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2019 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2020 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,9 +22,7 @@
 #include <sys/resource.h>
 #include <sys/socket.h>
 
-#if !defined(KORE_NO_TLS)
 #include <openssl/rand.h>
-#endif
 
 #include <fcntl.h>
 #include <grp.h>
@@ -33,6 +31,10 @@
 
 #include "kore.h"
 
+#if defined(KORE_USE_ACME)
+#include "acme.h"
+#endif
+
 #if !defined(KORE_NO_HTTP)
 #include "http.h"
 #endif
@@ -49,15 +51,19 @@
 #include "python_api.h"
 #endif
 
+#if defined(KORE_USE_CURL)
+#include "curl.h"
+#endif
+
+#if defined(__linux__)
+#include "seccomp.h"
+#endif
+
 #if !defined(WAIT_ANY)
 #define WAIT_ANY		(-1)
 #endif
 
-#if !defined(KORE_NO_TLS)
-#define WORKER_SOLO_COUNT	2
-#else
-#define WORKER_SOLO_COUNT	1
-#endif
+#define WORKER_SOLO_COUNT	3
 
 #define WORKER(id)						\
 	(struct kore_worker *)((u_int8_t *)kore_workers +	\
@@ -70,17 +76,14 @@ struct wlock {
 
 static int	worker_trylock(void);
 static void	worker_unlock(void);
+static void	worker_reaper(pid_t, int);
 
 static inline int	worker_acceptlock_obtain(void);
 static inline void	worker_acceptlock_release(void);
 static void		worker_accept_avail(struct kore_msg *, const void *);
 
-#if !defined(KORE_NO_TLS)
 static void	worker_entropy_recv(struct kore_msg *, const void *);
 static void	worker_keymgr_response(struct kore_msg *, const void *);
-static int	worker_keymgr_response_verify(struct kore_msg *, const void *,
-		    struct kore_domain **);
-#endif
 
 static int				accept_avail;
 static struct kore_worker		*kore_workers;
@@ -101,17 +104,15 @@ kore_worker_init(void)
 {
 	size_t			len;
 	struct kore_worker	*kw;
-	u_int16_t		i, cpu;
+	u_int16_t		idx, id, cpu;
 
 	worker_no_lock = 0;
 
 	if (worker_count == 0)
 		worker_count = cpu_count;
 
-#if !defined(KORE_NO_TLS)
-	/* account for the key manager. */
-	worker_count += 1;
-#endif
+	/* Account for the keymgr/acme even if we don't end up starting it. */
+	worker_count += 2;
 
 	len = sizeof(*accept_lock) +
 	    (sizeof(struct kore_worker) * worker_count);
@@ -137,29 +138,46 @@ kore_worker_init(void)
 	}
 
 	/* Setup log buffers. */
-	for (i = 0; i < worker_count; i++) {
-		kw = WORKER(i);
+	for (idx = KORE_WORKER_BASE; idx < worker_count; idx++) {
+		kw = WORKER(idx);
 		kw->lb.offset = 0;
 	}
 
-	cpu = 0;
-	for (i = 0; i < worker_count; i++) {
-		kore_worker_spawn(i, cpu++);
-		if (cpu == cpu_count)
+	/* Now start all the workers. */
+	id = 1;
+	cpu = 1;
+	for (idx = KORE_WORKER_BASE; idx < worker_count; idx++) {
+		if (cpu >= cpu_count)
 			cpu = 0;
+		kore_worker_spawn(idx, id++, cpu++);
+	}
+
+	if (keymgr_active) {
+#if defined(KORE_USE_ACME)
+		/* The ACME process is only started if we need it. */
+		if (acme_provider) {
+			kore_worker_spawn(KORE_WORKER_ACME_IDX,
+			    KORE_WORKER_ACME, 0);
+		}
+#endif
+
+		/* Now we can start the keymgr. */
+		kore_worker_spawn(KORE_WORKER_KEYMGR_IDX,
+		    KORE_WORKER_KEYMGR, 0);
 	}
 }
 
 void
-kore_worker_spawn(u_int16_t id, u_int16_t cpu)
+kore_worker_spawn(u_int16_t idx, u_int16_t id, u_int16_t cpu)
 {
 	struct kore_worker	*kw;
 
-	kw = WORKER(id);
+	kw = WORKER(idx);
 	kw->id = id;
 	kw->cpu = cpu;
 	kw->has_lock = 0;
 	kw->active_hdlr = NULL;
+	kw->running = 1;
 
 	if (socketpair(AF_UNIX, SOCK_STREAM, 0, kw->pipe) == -1)
 		fatal("socketpair(): %s", errno_s);
@@ -194,7 +212,7 @@ kore_worker_shutdown(void)
 	struct kore_worker	*kw;
 	pid_t			pid;
 	int			status;
-	u_int16_t		id, done;
+	u_int16_t		idx, done;
 
 	if (!kore_quiet) {
 		kore_log(LOG_NOTICE,
@@ -202,21 +220,37 @@ kore_worker_shutdown(void)
 	}
 
 	for (;;) {
-		for (id = 0; id < worker_count; id++) {
-			kw = WORKER(id);
+		for (idx = 0; idx < worker_count; idx++) {
+			kw = WORKER(idx);
+			if (kw->running == 0)
+				continue;
+
 			if (kw->pid != 0) {
 				pid = waitpid(kw->pid, &status, 0);
 				if (pid == -1)
 					continue;
+
+#if defined(__linux__)
+				kore_seccomp_trace(kw->pid, status);
+#endif
+
 				kw->pid = 0;
+				kw->running = 0;
+
+				if (!kore_quiet) {
+					kore_log(LOG_NOTICE, "worker %s exited",
+					    kore_worker_name(kw->id));
+				}
 			}
 		}
 
 		done = 0;
-		for (id = 0; id < worker_count; id++) {
-			kw = WORKER(id);
-			if (kw->pid == 0)
+		for (idx = 0; idx < worker_count; idx++) {
+			kw = WORKER(idx);
+			if (kw->running == 0) {
 				done++;
+				continue;
+			}
 		}
 
 		if (done == worker_count)
@@ -232,11 +266,11 @@ kore_worker_shutdown(void)
 void
 kore_worker_dispatch_signal(int sig)
 {
-	u_int16_t		id;
+	u_int16_t		idx;
 	struct kore_worker	*kw;
 
-	for (id = 0; id < worker_count; id++) {
-		kw = WORKER(id);
+	for (idx = 0; idx < worker_count; idx++) {
+		kw = WORKER(idx);
 		if (kill(kw->pid, sig) == -1) {
 			kore_debug("kill(%d, %d): %s", kw->pid, sig, errno_s);
 		}
@@ -306,31 +340,24 @@ kore_worker_privdrop(const char *runas, const char *root)
 			fatalx("cannot drop privileges");
 	}
 
-#if defined(KORE_USE_PLATFORM_PLEDGE)
-	kore_platform_pledge();
-#endif
-
+	kore_platform_sandbox();
 }
 
 void
 kore_worker_entry(struct kore_worker *kw)
 {
 	struct kore_runtime_call	*rcall;
-	char				buf[16];
-	int				quit, had_lock;
-	u_int64_t			netwait, now, next_prune;
-#if !defined(KORE_NO_TLS)
 	u_int64_t			last_seed;
-#endif
+	int				quit, had_lock;
+	u_int64_t			netwait, now, next_timeo;
 
 	worker = kw;
 
-	(void)snprintf(buf, sizeof(buf), "[wrk %d]", kw->id);
-#if !defined(KORE_NO_TLS)
-	if (kw->id == KORE_WORKER_KEYMGR)
-		(void)snprintf(buf, sizeof(buf), "[keymgr]");
+#if defined(__linux__)
+	kore_seccomp_traceme();
 #endif
-	kore_platform_proctitle(buf);
+
+	kore_platform_proctitle(kore_worker_name(kw->id));
 
 	if (worker_set_affinity == 1)
 		kore_platform_worker_setcpu(kw);
@@ -339,17 +366,27 @@ kore_worker_entry(struct kore_worker *kw)
 
 	kore_signal_setup();
 
-#if !defined(KORE_NO_TLS)
 	if (kw->id == KORE_WORKER_KEYMGR) {
 		kore_keymgr_run();
 		exit(0);
 	}
+
+#if defined(KORE_USE_ACME)
+	if (kw->id == KORE_WORKER_ACME) {
+		kore_acme_run();
+		exit(0);
+	}
 #endif
+
 	net_init();
 	kore_connection_init();
 	kore_platform_event_init();
 	kore_msg_worker_init();
 
+#if defined(KORE_USE_TASKS)
+	kore_task_init();
+#endif
+
 	kore_worker_privdrop(kore_runas_user, kore_root_path);
 
 #if !defined(KORE_NO_HTTP)
@@ -363,28 +400,28 @@ kore_worker_entry(struct kore_worker *kw)
 
 	quit = 0;
 	had_lock = 0;
-	next_prune = 0;
+	next_timeo = 0;
 	accept_avail = 1;
 	worker_active_connections = 0;
 
-#if defined(KORE_USE_PGSQL)
-	kore_pgsql_sys_init();
-#endif
-
-#if defined(KORE_USE_TASKS)
-	kore_task_init();
-#endif
-
-#if !defined(KORE_NO_TLS)
 	last_seed = 0;
-	kore_msg_register(KORE_MSG_CRL, worker_keymgr_response);
-	kore_msg_register(KORE_MSG_ENTROPY_RESP, worker_entropy_recv);
-	kore_msg_register(KORE_MSG_CERTIFICATE, worker_keymgr_response);
-	if (worker->restarted) {
-		kore_msg_send(KORE_WORKER_KEYMGR,
-		    KORE_MSG_CERTIFICATE_REQ, NULL, 0);
-	}
+
+	if (keymgr_active) {
+		kore_msg_register(KORE_MSG_CRL, worker_keymgr_response);
+		kore_msg_register(KORE_MSG_ENTROPY_RESP, worker_entropy_recv);
+		kore_msg_register(KORE_MSG_CERTIFICATE, worker_keymgr_response);
+
+		if (worker->restarted) {
+			kore_msg_send(KORE_WORKER_KEYMGR,
+			    KORE_MSG_CERTIFICATE_REQ, NULL, 0);
+		}
+#if defined(KORE_USE_ACME)
+		kore_msg_register(KORE_ACME_CHALLENGE_SET_CERT,
+		    worker_keymgr_response);
+		kore_msg_register(KORE_ACME_CHALLENGE_CLEAR_CERT,
+		    worker_keymgr_response);
 #endif
+	}
 
 	kore_msg_register(KORE_MSG_ACCEPT_AVAILABLE, worker_accept_avail);
 
@@ -409,17 +446,15 @@ kore_worker_entry(struct kore_worker *kw)
 	for (;;) {
 		now = kore_time_ms();
 
-#if !defined(KORE_NO_TLS)
-		if ((now - last_seed) > KORE_RESEED_TIME) {
+		if (keymgr_active && (now - last_seed) > KORE_RESEED_TIME) {
 			kore_msg_send(KORE_WORKER_KEYMGR,
 			    KORE_MSG_ENTROPY_REQ, NULL, 0);
 			last_seed = now;
 		}
-#endif
 
 		if (!worker->has_lock && accept_avail) {
-			accept_avail = 0;
 			if (worker_acceptlock_obtain()) {
+				accept_avail = 0;
 				if (had_lock == 0) {
 					kore_platform_enable_accept();
 					had_lock = 1;
@@ -429,8 +464,18 @@ kore_worker_entry(struct kore_worker *kw)
 
 		netwait = kore_timer_next_run(now);
 
-		if (netwait == KORE_WAIT_INFINITE && http_request_count > 0)
-			netwait = 100;
+		if (netwait == KORE_WAIT_INFINITE) {
+			if (sig_recv != 0)
+				netwait = 10;
+#if !defined(KORE_NO_HTTP)
+			if (http_request_count > 0)
+				netwait = 100;
+#endif
+#if defined(KORE_USE_PYTHON)
+			if (kore_python_coro_pending())
+				netwait = 10;
+#endif
+		}
 
 		kore_platform_event_wait(netwait);
 		now = kore_time_ms();
@@ -471,19 +516,22 @@ kore_worker_entry(struct kore_worker *kw)
 			break;
 
 		kore_timer_run(now);
-
+#if defined(KORE_USE_CURL)
+		kore_curl_run_scheduled();
+		kore_curl_do_timeout();
+#endif
 #if !defined(KORE_NO_HTTP)
 		http_process();
 #endif
 #if defined(KORE_USE_PYTHON)
 		kore_python_coro_run();
 #endif
-
-		if (next_prune <= now) {
+		if (next_timeo <= now) {
 			kore_connection_check_timeout(now);
-			kore_connection_prune(KORE_CONNECTION_PRUNE_DISCONNECT);
-			next_prune = now + 500;
+			next_timeo = now + 500;
 		}
+
+		kore_connection_prune(KORE_CONNECTION_PRUNE_DISCONNECT);
 	}
 
 	rcall = kore_runtime_getcall("kore_worker_teardown");
@@ -492,6 +540,9 @@ kore_worker_entry(struct kore_worker *kw)
 		kore_free(rcall);
 	}
 
+	kore_msg_send(KORE_MSG_PARENT, KORE_MSG_SHUTDOWN, NULL, 0);
+	kore_server_cleanup();
+
 	kore_platform_event_cleanup();
 	kore_connection_cleanup();
 	kore_domain_cleanup();
@@ -518,10 +569,7 @@ kore_worker_entry(struct kore_worker *kw)
 void
 kore_worker_reap(void)
 {
-	u_int16_t		id;
 	pid_t			pid;
-	struct kore_worker	*kw;
-	const char		*func;
 	int			status;
 
 	for (;;) {
@@ -537,14 +585,99 @@ kore_worker_reap(void)
 			return;
 		}
 
-		break;
-	}
+		if (pid == 0)
+			return;
 
-	if (pid == 0)
+		worker_reaper(pid, status);
+	}
+}
+
+void
+kore_worker_make_busy(void)
+{
+	if (worker_count == WORKER_SOLO_COUNT || worker_no_lock == 1)
 		return;
 
-	for (id = 0; id < worker_count; id++) {
-		kw = WORKER(id);
+	if (worker->has_lock) {
+		worker_unlock();
+		worker->has_lock = 0;
+		kore_msg_send(KORE_MSG_WORKER_ALL,
+		    KORE_MSG_ACCEPT_AVAILABLE, NULL, 0);
+	}
+}
+
+int
+kore_worker_keymgr_response_verify(struct kore_msg *msg, const void *data,
+    struct kore_domain **out)
+{
+	struct kore_server		*srv;
+	struct kore_domain		*dom;
+	const struct kore_x509_msg	*req;
+
+	if (msg->length < sizeof(*req)) {
+		kore_log(LOG_WARNING,
+		    "short keymgr message (%zu)", msg->length);
+		return (KORE_RESULT_ERROR);
+	}
+
+	req = (const struct kore_x509_msg *)data;
+	if (msg->length != (sizeof(*req) + req->data_len)) {
+		kore_log(LOG_WARNING,
+		    "invalid keymgr payload (%zu)", msg->length);
+		return (KORE_RESULT_ERROR);
+	}
+
+	if (req->domain[KORE_DOMAINNAME_LEN] != '\0') {
+		kore_log(LOG_WARNING, "domain not NUL-terminated");
+		return (KORE_RESULT_ERROR);
+
+	}
+
+	if (out == NULL)
+		return (KORE_RESULT_OK);
+
+	dom = NULL;
+
+	LIST_FOREACH(srv, &kore_servers, list) {
+		dom = NULL;
+
+		if (srv->tls == 0)
+			continue;
+
+		TAILQ_FOREACH(dom, &srv->domains, list) {
+			if (!strcmp(dom->domain, req->domain))
+				break;
+		}
+
+		if (dom != NULL)
+			break;
+	}
+
+	if (dom == NULL) {
+		kore_log(LOG_WARNING,
+		    "got keymgr response for domain that does not exist");
+		return (KORE_RESULT_ERROR);
+	}
+
+	*out = dom;
+
+	return (KORE_RESULT_OK);
+}
+
+static void
+worker_reaper(pid_t pid, int status)
+{
+	u_int16_t		idx;
+	struct kore_worker	*kw;
+	const char		*func;
+
+#if defined(__linux__)
+	if (kore_seccomp_trace(pid, status))
+		return;
+#endif
+
+	for (idx = 0; idx < worker_count; idx++) {
+		kw = WORKER(idx);
 		if (kw->pid != pid)
 			continue;
 
@@ -554,6 +687,13 @@ kore_worker_reap(void)
 			    kw->id, pid, status);
 		}
 
+		kw->running = 0;
+
+		if (WIFEXITED(status) && WEXITSTATUS(status) == 0) {
+			kw->pid = 0;
+			break;
+		}
+
 		func = "none";
 #if !defined(KORE_NO_HTTP)
 		if (kw->active_hdlr != NULL)
@@ -563,9 +703,17 @@ kore_worker_reap(void)
 		    "worker %d (pid: %d) (hdlr: %s) gone",
 		    kw->id, kw->pid, func);
 
-#if !defined(KORE_NO_TLS)
-		if (id == KORE_WORKER_KEYMGR) {
-			kore_log(LOG_CRIT, "keymgr gone, stopping");
+#if defined(__linux__)
+		if (WIFSIGNALED(status) && WTERMSIG(status) == SIGSYS) {
+			kore_log(LOG_NOTICE,
+			    "worker %d died from sandbox violation", kw->id);
+		}
+#endif
+
+		if (kw->id == KORE_WORKER_KEYMGR ||
+		    kw->id == KORE_WORKER_ACME) {
+			kore_log(LOG_CRIT,
+			    "keymgr or acme process gone, stopping");
 			kw->pid = 0;
 			if (raise(SIGTERM) != 0) {
 				kore_log(LOG_WARNING,
@@ -573,7 +721,6 @@ kore_worker_reap(void)
 			}
 			break;
 		}
-#endif
 
 		if (kw->pid == accept_lock->current &&
 		    worker_no_lock == 0)
@@ -603,27 +750,13 @@ kore_worker_reap(void)
 		kore_log(LOG_NOTICE, "restarting worker %d", kw->id);
 		kw->restarted = 1;
 		kore_msg_parent_remove(kw);
-		kore_worker_spawn(kw->id, kw->cpu);
+		kore_worker_spawn(idx, kw->id, kw->cpu);
 		kore_msg_parent_add(kw);
 
 		break;
 	}
 }
 
-void
-kore_worker_make_busy(void)
-{
-	if (worker_count == WORKER_SOLO_COUNT || worker_no_lock == 1)
-		return;
-
-	if (worker->has_lock) {
-		worker_unlock();
-		worker->has_lock = 0;
-		kore_msg_send(KORE_MSG_WORKER_ALL,
-		    KORE_MSG_ACCEPT_AVAILABLE, NULL, 0);
-	}
-}
-
 static inline void
 worker_acceptlock_release(void)
 {
@@ -701,7 +834,7 @@ worker_unlock(void)
 {
 	accept_lock->current = 0;
 	if (!__sync_bool_compare_and_swap(&(accept_lock->lock), 1, 0))
-		kore_log(LOG_NOTICE, "worker_unlock(): wasnt locked");
+		kore_log(LOG_NOTICE, "worker_unlock(): wasn't locked");
 }
 
 static void
@@ -710,7 +843,6 @@ worker_accept_avail(struct kore_msg *msg, const void *data)
 	accept_avail = 1;
 }
 
-#if !defined(KORE_NO_TLS)
 static void
 worker_entropy_recv(struct kore_msg *msg, const void *data)
 {
@@ -730,65 +862,45 @@ worker_keymgr_response(struct kore_msg *msg, const void *data)
 	struct kore_domain		*dom;
 	const struct kore_x509_msg	*req;
 
-	if (!worker_keymgr_response_verify(msg, data, &dom))
+	if (!kore_worker_keymgr_response_verify(msg, data, &dom))
 		return;
 
 	req = (const struct kore_x509_msg *)data;
 
 	switch (msg->id) {
 	case KORE_MSG_CERTIFICATE:
-		kore_domain_tlsinit(dom, req->data, req->data_len);
+		kore_domain_tlsinit(dom, KORE_PEM_CERT_CHAIN,
+		    req->data, req->data_len);
 		break;
 	case KORE_MSG_CRL:
 		kore_domain_crl_add(dom, req->data, req->data_len);
 		break;
+#if defined(KORE_USE_ACME)
+	case KORE_ACME_CHALLENGE_SET_CERT:
+		if (dom->ssl_ctx == NULL) {
+			kore_domain_tlsinit(dom, KORE_DER_CERT_DATA,
+			    req->data, req->data_len);
+		}
+
+		kore_free(dom->acme_cert);
+		dom->acme_cert_len = req->data_len;
+		dom->acme_cert = kore_calloc(1, req->data_len);
+		memcpy(dom->acme_cert, req->data, req->data_len);
+
+		kore_log(LOG_NOTICE, "[%s] tls-alpn-01 challenge active",
+		    dom->domain);
+		dom->acme_challenge = 1;
+		break;
+	case KORE_ACME_CHALLENGE_CLEAR_CERT:
+		dom->acme_cert_len = 0;
+		dom->acme_challenge = 0;
+		kore_free(dom->acme_cert);
+		kore_log(LOG_NOTICE, "[%s] tls-alpn-01 challenge disabled",
+		    dom->domain);
+		break;
+#endif
 	default:
 		kore_log(LOG_WARNING, "unknown keymgr request %u", msg->id);
 		break;
 	}
 }
-
-static int
-worker_keymgr_response_verify(struct kore_msg *msg, const void *data,
-    struct kore_domain **out)
-{
-	struct kore_domain		*dom;
-	const struct kore_x509_msg	*req;
-
-	if (msg->length < sizeof(*req)) {
-		kore_log(LOG_WARNING,
-		    "short keymgr message (%zu)", msg->length);
-		return (KORE_RESULT_ERROR);
-	}
-
-	req = (const struct kore_x509_msg *)data;
-	if (msg->length != (sizeof(*req) + req->data_len)) {
-		kore_log(LOG_WARNING,
-		    "invalid keymgr payload (%zu)", msg->length);
-		return (KORE_RESULT_ERROR);
-	}
-
-	if (req->domain_len > KORE_DOMAINNAME_LEN) {
-		kore_log(LOG_WARNING,
-		    "invalid keymgr domain (%u)",
-		    req->domain_len);
-		return (KORE_RESULT_ERROR);
-	}
-
-	dom = NULL;
-	TAILQ_FOREACH(dom, &domains, list) {
-		if (!strncmp(dom->domain, req->domain, req->domain_len))
-			break;
-	}
-
-	if (dom == NULL) {
-		kore_log(LOG_WARNING,
-		    "got keymgr response for domain that does not exist");
-		return (KORE_RESULT_ERROR);
-	}
-
-	*out = dom;
-
-	return (KORE_RESULT_OK);
-}
-#endif
